from abc import abstractmethod

from tradingkit.pubsub.core.event import Event
from tradingkit.pubsub.core.publisher import Publisher
from tradingkit.pubsub.core.subscriber import Subscriber


class Adapter(Subscriber, Publisher):

    @abstractmethod
    def adapt(self, event: Event) -> Event:
        pass

    def on_event(self, event: Event):
        self.dispatch(self.adapt(event))



from tradingkit.exchange.testex import TestEX


class Runner:

    @staticmethod
    def run(feeder, exchange, plotter, strategy, bridge, feeder_adapters=None, optimize=False, plot=False):
        if feeder_adapters is None:
            feeder_adapters = []

        chain = feeder
        for adapter in feeder_adapters:
            chain.register(adapter)
            chain = adapter

        bridge.register(strategy)
        if plot:
            bridge.register(plotter)
            strategy.register(plotter)

        if isinstance(exchange, TestEX):
            chain.register(exchange)
            chain = exchange
        chain.register(bridge)
        feeder.feed()
        if optimize:
            result = strategy.finish(False)
        else:
            result = strategy.finish()
            print("Trading results")
            for info in result:
                print("%15s: %10.2f" % (info, result[info]))

        if plot:
            plotter.plot()

        return result



from tradingkit.data.adapter.adapter import Adapter
from tradingkit.pubsub.core.event import Event
from tradingkit.pubsub.event.trade import Trade


class BitmexNormalizer(Adapter):

    def __init__(self, symbol):
        super().__init__()
        self.symbol = symbol

    def adapt(self, event: Event) -> Event:
        trade = event.payload.copy()
        trade['symbol'] = self.symbol
        trade['cost'] = trade['amount']
        trade['amount'] = trade['amount'] / trade['price']
        return Trade(trade)

    def subscribed_events(self) -> list:
        return [Trade]



import logging
from abc import abstractmethod

from ccxt import Exchange

from tradingkit.pubsub.core.event import Event
from tradingkit.strategy.strategy import Strategy
from tradingkit.strategy.strategy_event import StrategyEvent


class StateMachineStrategy(Strategy):

    def __init__(self, exchange: Exchange, config=None):
        super().__init__(exchange, config)
        self.state = None
        self.prev = None

    def start(self):
        super().start()
        self.state = self.recover_state() or self.initial_state()

    def on_event(self, event: Event):
        logging.debug("Event: %s: %s" % (event.__class__.__name__, event.payload))
        super().on_event(event)
        if self.prev != self.state:
            logging.debug("State change: %s -> %s" % (self.prev, self.state))
            self.prev = self.state
        self.state = self.state.on_event(StrategyEvent(event, self))

    def recover_state(self):
        return None

    @abstractmethod
    def initial_state(self):
        pass



import logging
from abc import ABC, abstractmethod

from ccxt import Exchange

from tradingkit.pubsub.core.event import Event
from tradingkit.pubsub.core.publisher import Publisher
from tradingkit.pubsub.core.subscriber import Subscriber


class Strategy(Publisher, Subscriber, ABC):

    def __init__(self, exchange: Exchange, config=None):
        super().__init__()
        self.exchange = exchange
        self.config = config
        self.start_equity = None
        self.is_started = False
        self.start_base_balance = None
        self.start_base_equity = None

    @abstractmethod
    def get_symbol(self):
        pass

    def start(self):
        logging.info("Start strategy %s" % str(self.__class__))
        balance = self.exchange.fetch_balance()['total']
        symbol = self.config['symbol']
        base, quote = symbol.split('/')
        quote_balance = balance[quote] if quote in balance else 0
        self.start_equity = quote_balance + balance[base] * self.exchange.fetch_ticker(symbol)['bid']
        self.start_base_equity = quote_balance / self.exchange.fetch_ticker(symbol)['bid'] + balance[base]
        self.start_base_balance = balance[base]
        logging.info("Initial Equity: %s" % str(self.start_equity))

    def on_event(self, event: Event):
        if not self.is_started:
            self.is_started = True
            self.start()

    def finish(self, show_info=True):
        logging.info("Finish strategy %s" % str(self.__class__))
        balance = self.exchange.fetch_balance()['total']
        symbol = self.config['symbol']
        base, quote = symbol.split('/')
        price = self.exchange.fetch_ticker(symbol)['bid']
        end_equity = balance[quote] + balance[base] * price
        end_base_equity = balance[quote] / price + balance[base]
        try:
            position = self.exchange.private_get_position()[0]
            pnl = (price / position['avgEntryPrice'] - 1) * position['currentQty']
        except Exception as e:
            pnl = 0
        end_equity += pnl
        logging.info("Final Balances: %s" % str(balance))
        logging.info("Equity: %.2f EUR" % end_equity)

        return {
            "start_equity": self.start_equity,
            "end_equity": end_equity,
            "profit": end_equity - self.start_equity,
            "profit_percent": (end_equity - self.start_equity) / self.start_equity * 100.0,
            "quote_balance": int(balance[quote]),
            "base_balance": balance[base],
            "max_drawdown": self.exchange.get_max_draw_down(),
            "sharpe_ratio": self.exchange.get_sharpe_ratio(),
            "start_base_balance": self.start_base_balance,
            "end_base_equity": end_base_equity
        }



#!/usr/bin/env python


import asyncio
import json
import logging
import multiprocessing

import websockets


class PlotServer:

    def __init__(self, port):
        self.port = port
        self.DATA = {'price': {"OHLC": [],
                               "close": [],
                               "volume": [],
                               "pos_vol": [],
                               "equity": [],
                               "base_equity": [],
                               "invested": [],
                               "hold": [],
                               "pos_p": []
                               },
                     'buy': [], 'sell': [], 'assets': []}
        self.NEW_DATA = {}

        self.USERS = set()

    def state_event(self):
        return json.dumps({"type": "start", "data": self.DATA})

    def update_event(self):
        return json.dumps({"type": self.NEW_DATA['type'], "data": self.NEW_DATA['data']})

    async def notify_state(self):
        if self.USERS:  # asyncio.wait doesn't accept an empty list
            message = self.state_event()
            await asyncio.wait([user.send(message) for user in self.USERS])

    async def notify(self):
        if self.USERS:  # asyncio.wait doesn't accept an empty list
            message = self.update_event()
            await asyncio.wait([user.send(message) for user in self.USERS])

    async def register(self, websocket):
        self.USERS.add(websocket)

    async def unregister(self, websocket):
        self.USERS.remove(websocket)

    async def update(self, websocket, path):
        try:
            async for message in websocket:
                payload = json.loads(message)
                if payload["action"] == "produce":
                    if payload['type'] == "price":
                        for key in payload['data'].keys():
                            # print(key)
                            self.DATA['price'][key].append(payload['data'][key])
                        self.NEW_DATA['data'] = payload['data']
                        self.NEW_DATA['type'] = payload['type']
                    elif payload['type'] == "buy":
                        self.NEW_DATA['data'] = payload['data']
                        self.NEW_DATA['type'] = payload['type']
                        self.DATA['buy'].append(payload['data'])
                    elif payload['type'] == "sell":
                        self.NEW_DATA['data'] = payload['data']
                        self.NEW_DATA['type'] = payload['type']
                        self.DATA['sell'].append(payload['data'])
                    else:
                        self.DATA[payload['type']].append(payload['data'])

                    await self.notify()
                elif payload["action"] == "start":
                    await self.notify_state()
                elif payload["action"] == "register":
                    await self.register(websocket)
                else:
                    logging.error("unsupported event: {}", payload)
        finally:
            if websocket in self.USERS:
                await self.unregister(websocket)

    def run(self):
        start_server = websockets.serve(self.update, "localhost", self.port)
        asyncio.get_event_loop().run_until_complete(start_server)
        asyncio.get_event_loop().run_forever()


# ps = PlotServer()
# ps.run()
# print("running...")

if __name__ == '__main__':
    ps = PlotServer()
    p = multiprocessing.Process(target=ps.run)
    p.start()
    print("running...")



import logging
import threading
import time

from tradingkit.data.feed.feeder import Feeder
from tradingkit.pubsub.core.event import Event
from tradingkit.pubsub.core.publisher import Publisher
from tradingkit.pubsub.core.subscriber import Subscriber
from tradingkit.pubsub.event.book import Book
from tradingkit.pubsub.event.candle import Candle
from tradingkit.pubsub.event.order import Order
from tradingkit.pubsub.event.position import Position
from tradingkit.pubsub.event.trade import Trade


class AggregatorFeeder(Feeder, Subscriber, Publisher):

    def __init__(self, feeders):
        super().__init__()
        self.feeders = feeders

    def subscribed_events(self) -> list:
        return [Order, Book, Trade, Candle, Position]

    def on_event(self, event: Event):
        self.dispatch(event)

    def feed(self):
        children = []
        for feeder in self.feeders:
            feeder.register(self)
            child = threading.Thread(target=feeder.feed)
            child.start()
            children.append(child)

        # Active loop to check feeder children
        while True:
            time.sleep(10)
            for child in children:
                if not child.is_alive():
                    raise ValueError("Feeder thread %s ended" % child.getName())






from tradingkit.data.adapter.adapter import Adapter
from tradingkit.pubsub.core.event import Event
from tradingkit.pubsub.event.trade import Trade


class OutlierTradeFilter(Adapter):
    """
    Detects anomalies based on last trade price
    Takes the trade and checks if price changed more than the :max_rate permitted
    If anomaly is detected, it returns the last trade whithout bias.
    """

    def __init__(self, max_rate):
        super().__init__()
        self.max_rate = max_rate
        self.last_trade = None

    def adapt(self, event: Event) -> Event:
        trade = event.payload
        if self.last_trade is None:
            self.last_trade = trade
            return event

        change_rate = abs(trade['price'] - self.last_trade['price']) / self.last_trade['price']
        if change_rate > self.max_rate:
            corrected_trade = self.last_trade.copy()
            corrected_trade['timestamp'] = trade['timestamp']
            return Trade(corrected_trade)

        self.last_trade = trade
        return event

    def subscribed_events(self) -> list:
        return [Trade]



import telebot


class TelegramAPI:

    def __init__(self, bot_token, bot_chat_id):
        self.bot_token = bot_token
        self.bot_chatID = bot_chat_id
        self.tb = telebot.TeleBot(self.bot_token)

    def send_text(self, bot_message):
        self.tb.send_message(self.bot_chatID, bot_message)

    def send_image(self, route):
        photo = open(route, 'rb')
        self.tb.send_photo(self.bot_chatID, photo)



from __future__ import annotations

import json
from abc import abstractmethod

from tradingkit.state_machine.state import State
from tradingkit.strategy.strategy_event import StrategyEvent


class StrategyState(State):

    @abstractmethod
    def on_event(self, strategy_event: StrategyEvent) -> StrategyState:
        pass

    def serialize(self) -> str:
        return json.dumps(self)

    @staticmethod
    def unserialize(data: str):
        return json.loads(data)



import sys
import time

from ccxt import DDoSProtection, RequestTimeout, Exchange

from tradingkit.data.fetch.fetcher import Fetcher
from datetime import datetime


class CCXTFetcher(Fetcher):
    def __init__(self, exchange: Exchange):
        self.exchange = exchange
        self.wait_seconds = 60

        self.candles = None
        self.last_candle = None
        self.timeframes = {"1m": 60,
                           "5m": 300,
                           "15m": 900,
                           "30m": 1800,
                           "1H": 3600,
                           "4H": 14400,
                           "6H": 21600,
                           "12H": 43200,
                           "1D": 86400,
                           "2D": 172800,
                           "4D": 345600,
                           "1W": 604800,
                           "2W": 1209600,
                           "1_M": 2629800}

    def fetch(self, symbol, since8601=None, to8601=None):
        since = 0 if since8601 is None else self.exchange.parse8601(since8601)
        to = self.exchange.milliseconds() if to8601 is None else self.exchange.parse8601(to8601)
        end = False
        while since < to and not end:
            try:
                sys.stdout.write("Since %s\n" % (time.ctime(since // 1000)))
                step = self.exchange.fetch_trades(symbol, since)
                if len(step) > 0:
                    if since != step[-1]['timestamp']:
                        since = step[-1]['timestamp']
                        yield step
                    else:
                        sys.stdout.write("More trades in one millisecond than response length\n")
                        since += 1
                else:
                    sys.stdout.write("End of trades\n")
                    end = True
            except DDoSProtection:
                sys.stderr.write("Api call rate limit reached, waiting %d seconds...\n" % self.wait_seconds)
                time.sleep(self.wait_seconds)
            except RequestTimeout:
                sys.stderr.write("Request timeout, retrying...\n")

    def fetch_all(self, symbol, since8601, to8601=None):
        trades = []
        last_trade = None
        for step in self.fetch(symbol, since8601, to8601):
            for trade in step:
                if last_trade is None:
                    last_trade = trade.copy()
                elif last_trade['price'] == trade['price'] and last_trade['side'] == trade['side']:
                    last_trade['amount'] += trade['amount']
                    # last_trade['cost'] += trade['cost']  # on some exchanges is None
                else:
                    trades.append(last_trade)
                    self.trade_to_candle(last_trade)
                    last_trade = trade.copy()
            trades.append(last_trade)
            print("ACC trades", len(trades))
        return [trades, self.candles]

    def trade_to_candle(self, trade):

        if self.candles is None:
            self.candles = {x:{} for x in self.timeframes.keys()}

        if self.last_candle is None:
            self.last_candle = {x:None for x in self.timeframes.keys()}

        for tf in self.timeframes.keys():
            sec = self.timeframes[tf]
            key = trade['timestamp'] // (sec * 1000) * sec
            key = str(datetime.fromtimestamp(key))
            if self.last_candle[tf] is not None and key in self.candles[tf]:
                self.candles[tf][key]['high'] = max(self.candles[tf][key]['high'], trade['price'])
                self.candles[tf][key]['low'] = min(self.candles[tf][key]['low'], trade['price'])
                self.candles[tf][key]['close'] = trade['price']
                self.candles[tf][key]['vol'] += trade['amount']
                self.candles[tf][key]['cost'] += trade['cost']
                self.candles[tf][key]['trades'] += 1
            else:
                self.candles[tf][key] = {
                    'datetime': key,
                    'open': trade['price'],
                    'high': trade['price'],
                    'low': trade['price'],
                    'close': trade['price'],
                    'vol': trade['amount'],
                    'cost': trade['cost'],
                    'trades': 1,
                    'timeframe': tf
                }
            self.last_candle[tf] = self.candles[tf][key]


import json
import logging
from datetime import datetime
import csv
from dateutil.relativedelta import relativedelta
from dateutil import parser

from tradingkit.utils.system import System
from tradingkit.data.feed.backtest_feeder import BacktestFeeder
from tradingkit.pubsub.core.publisher import Publisher
from tradingkit.pubsub.event.trade import Trade
from tradingkit.pubsub.event.funding import Funding


class FundingBacktestFeeder(BacktestFeeder, Publisher):
    def __init__(self, exchange='kraken', symbol='BTC/EUR', since8601=None, to8601=None):
        super().__init__(exchange, symbol, since8601, to8601)
        self.founding_rate_data = None
        self.next_founding_rate_index = 0

    def dispatch_month(self, exchange, symbol, year, month, since, to):
        import_dir = System.get_import_dir()
        base, quote = symbol.split('/')
        full_filename = '%s/%s-%s_%s-%d-%02d.json' % (import_dir, exchange, base, quote, year, month)

        for trade in json.load(open(full_filename, 'r')):
            if since.timestamp() <= trade['timestamp'] / 1000 < to.timestamp():
                self.dispatch_founding_rate(trade, exchange, base, quote)
                self.dispatch(Trade(trade))

    def dispatch_founding_rate(self, trade, exchange, base, quote):
        """Updates the founding rate"""
        if self.founding_rate_data is None:
            import_dir = System.get_import_dir()
            full_filename = '%s/funding_bitmex-%s_%s.json' % (import_dir, base, quote)

            self.founding_rate_data = []
            first_rate_timestamp = trade['timestamp'] - 8 * 60 * 60 * 1000
            for row in json.load(open(full_filename, 'r')):
                rate = float(row['fundingRate'])
                timestamp = int(parser.isoparse(row['timestamp']).timestamp() * 1000)
                if timestamp > first_rate_timestamp:
                    self.founding_rate_data.append({"timestamp": timestamp,
                                                    "rate": rate,
                                                    "date": row['timestamp'],
                                                    "price": trade['price'],
                                                    "symbol": trade['symbol']})
            self.dispatch(Funding(self.founding_rate_data[0]))
            self.next_founding_rate_index += 1

        elif self.next_founding_rate_index < len(self.founding_rate_data) and \
                trade['timestamp'] > self.founding_rate_data[self.next_founding_rate_index]["timestamp"]:
            self.dispatch(Funding(self.founding_rate_data[self.next_founding_rate_index]))
            self.next_founding_rate_index += 1








import base64
import hashlib
import hmac
import json
import time
import urllib

from ccxt import AuthenticationError
from tradingkit.data.feed.websocket_feeder import WebsocketFeeder
from tradingkit.pubsub.event.order import Order


class PrivateKrakenFeeder(WebsocketFeeder):

    # Converts symbols from normal to kraken vocab
    denormalized_symbol = {
        "BTC/EUR": "XBT/EUR",
        "BTC/USD": "XBT/USD",
        "BTC/USDT": "XBT/USDT",
    }

    # Converts symbols from kraken to normal vocab
    normalized_symbol = {
        "XBT/EUR": "BTC/EUR",
        "XBT/USD": "BTC/USD",
        "XBT/USDT": "BTC/USDT",
    }

    def __init__(self, symbol, credentials):
        super().__init__(symbol, credentials, "wss://ws-auth.kraken.com")

    def on_open(self, ws):
        token = self.get_ws_auth_token()
        ws.send(json.dumps({
            'event': 'subscribe',
            'subscription': {'name': 'ownTrades'},
            'snapshot': False,
            'pair': [self.denormalized_symbol[self.symbol]],
            'token': token
        }))

    def get_ws_auth_token(self):
        api_nonce = bytes(str(int(time.time() * 1000)), "utf-8")
        api_request = urllib.request.Request(
            "https://api.kraken.com/0/private/GetWebSocketsToken",
            b"nonce=%s" % api_nonce
        )
        api_request.add_header("API-Key", self.credentials['apiKey'])
        api_request.add_header("API-Sign", base64.b64encode(
            hmac.new(
                base64.b64decode(self.credentials['secret']),
                b"/0/private/GetWebSocketsToken" + hashlib.sha256(api_nonce + b"nonce=%s" % api_nonce).digest(),
                hashlib.sha512
            ).digest()
        ))
        resp = json.loads(urllib.request.urlopen(api_request).read())
        if 'result' in resp and 'token' in resp['result']:
            return resp['result']['token']
        else:
            raise AuthenticationError("Failed to authenticate with WebSocket: %s" % resp)

    def on_message(self, ws, message):
        data = json.loads(message)
        if "ownTrades" in data:
            for dict in data[0]:
                for order in dict:
                    order_data = {
                        'id': dict[order]['ordertxid'],
                        'timestamp': int(float(dict[order]['time']) * 1000),
                        'lastTradeTimestamp': int(float(dict[order]['time']) * 1000),
                        'status': 'filled',
                        'symbol': self.normalized_symbol[dict[order]['pair']],
                        'type': dict[order]['ordertype'],
                        'side': dict[order]['type'],
                        'price': float(dict[order]['price']),
                        'amount': float(dict[order]['vol'])
                    }
                    self.dispatch(Order(order_data))



from tradingkit.strategy.strategy import Strategy


class NoneStrategy(Strategy):

    def get_symbol(self):
        return 'BTC/USD'

    def subscribed_events(self) -> list:
        return []

    def finish(self):
        return {}



import logging
import uuid
from datetime import datetime

from ccxt import Exchange, OrderNotFound, InsufficientFunds

from tradingkit.pubsub.core.event import Event
from tradingkit.pubsub.core.publisher import Publisher
from tradingkit.pubsub.event.book import Book
from tradingkit.pubsub.core.subscriber import Subscriber
from tradingkit.pubsub.event.candle import Candle
from tradingkit.pubsub.event.open_order import OpenOrder
from tradingkit.pubsub.event.order import Order
from tradingkit.pubsub.event.plot import Plot
from tradingkit.pubsub.event.trade import Trade
from tradingkit.pubsub.event.funding import Funding


class TestEX(Publisher, Subscriber, Exchange):

    def __init__(self, params=None):
        super().__init__()
        if params is None:
            params = {
                'balance': {'EUR': 100000, 'BTC': 0},
                'fees': {
                    'maker': 0.002,
                    'taker': 0.002
                }
            }
        self.fees = params['fees'].copy()
        self.balance = params['balance'].copy()
        self.initial = params['balance'].copy()

        self.open_orders = {}
        self.history = []
        self.orderbooks = {}
        self.orders_scheduled_to_close = []
        self.max_invested = 0
        self.ready = False
        self.undispatched_orders = []

        # timestamp in milliseconds
        self.timestamp = 0

        #self.candles = {}
        #self.last_candle = None

    def sec(self):
        return self.seconds()

    def msec(self):
        return self.milliseconds()

    def usec(self):
        return self.microseconds()

    def seconds(self):
        return self.milliseconds() / 1000

    def milliseconds(self):
        return self.timestamp

    def microseconds(self):
        return self.milliseconds() * 1000

    def sleep(self, milliseconds):
        raise NotImplementedError("sleep() not implemented yet.")

    def fetch_open_orders(self, symbol=None, since=None, limit=None, params={}):
        return list(self.open_orders.values())

    def create_order(self, symbol, type, side, amount, price=None, params={}):
        order_id = str(uuid.uuid4())
        now = self.milliseconds()
        order = {
            'id': order_id,
            'timestamp': now,
            'status': 'closed' if type == 'market' else 'open',
            'symbol': symbol,
            'type': type,
            'side': side,
            'amount': amount,
            'price': price,
            'params': params
        }
        logging.debug("%s::create_order(%s)" % (self.__class__.__name__, order))
        if not self.funds_enough(order):
            raise InsufficientFunds("No funds enough: balance: %s, order: %s" % (str(self.balance), str(order)))
        self.open_orders[order_id] = order
        if type == 'market':
            base, quote = symbol.split('/')
            _price = self.orderbooks[symbol]['bids'][0][0]
            self.match_order({"price": _price, 'timestamp': now}, order, _price, base, quote)
            self.clean_orders()
        if order['status'] == 'filled':
            order['status'] = 'closed'
        else:
            self.dispatch(OpenOrder(order))
        return order

    def cancel_order(self, order_id, symbol=None, params={}):
        logging.debug("%s::cancel_order(%s)" % (self.__class__.__name__, order_id))
        if order_id in self.open_orders:
            order = self.open_orders[order_id]
            del self.open_orders[order_id]
            order['status'] = 'canceled'
            order['timestamp'] = self.milliseconds()
            self.dispatch(OpenOrder(order))
        else:
            raise OrderNotFound("Cannot cancel order: not found")

    def subscribed_events(self) -> list:
        return [Order, Trade, Book, Candle, Funding]

    def on_event(self, event: Event):
        if isinstance(event, Trade):
            trade = event.payload
            self.timestamp = trade['timestamp']
            self.update_order_book(trade)
            symbol = trade['symbol']
            assets = symbol.split('/')
            for asset in assets:
                if asset not in self.balance:
                    self.balance[asset] = 0
            base, quote = assets
            self.plot_balances(base, quote, trade['price'])
            for order_id in self.open_orders:
                order = self.open_orders[order_id]
                if order['symbol'] == symbol:
                    price = trade['price'] if order['type'] == 'market' else order['price']
                    self.match_order(trade, order, price, base, quote)
            for order in self.undispatched_orders:
                self.dispatch(Order(order))
            self.undispatched_orders = []
            self.clean_orders()
            self.ready = True
            book = self.orderbooks[symbol].copy()
            book['symbol'] = symbol
            self.dispatch(Book(book))
        self.dispatch(event)

    def dispatch(self, event: Event):
        # prevent from dispatching events if the exchange is not ready
        if self.ready:
            super().dispatch(event)

    def update_order_book(self, trade):
        symbol = trade['symbol']
        if symbol not in self.orderbooks:
            self.orderbooks[symbol] = {'bids': [], 'asks': []}
        self.orderbooks[symbol]['asks'] = [[trade['price'], trade['amount']]]
        self.orderbooks[symbol]['bids'] = [[trade['price'], trade['amount']]]
        self.orderbooks[symbol]['timestamp'] = trade['timestamp']

    def match_order(self, trade, order, price, base, quote):
        taker_or_maker = 'maker' if order['type'] == 'limit' else 'taker'
        fee = self.fees[taker_or_maker]
        if order['side'] == 'buy':
            if order['type'] == 'stop':
                if price <= trade['price']:
                    self.balance[quote] -= order['amount'] * price * (1 + fee)
                    self.balance[base] += order['amount']
                    order['status'] = 'filled'
            elif price >= trade['price']:
                self.balance[quote] -= order['amount'] * price * (1 + fee)
                self.balance[base] += order['amount']
                order['status'] = 'filled'
        else:
            if order['type'] == 'stop':
                if price >= trade['price']:
                    self.balance[base] -= order['amount']
                    self.balance[quote] += order['amount'] * price * (1 - fee)
                    order['status'] = 'filled'
            elif price <= trade['price']:
                self.balance[base] -= order['amount']
                self.balance[quote] += order['amount'] * price * (1 - fee)
                order['status'] = 'filled'
        if order['status'] == 'filled':
            order['lastTradeTimestamp'] = trade['timestamp']
            if order['price'] is None:
                order['price'] = price
            self.orders_scheduled_to_close.append(order)

    def fetch_l2_order_book(self, symbol, limit=None, params={}):
        return self.orderbooks[symbol]

    def clean_orders(self):
        orders_to_dispatch = []
        for order in self.orders_scheduled_to_close:
            del self.open_orders[order['id']]
            self.history.append(order)
            if order['type'] == 'market':
                self.undispatched_orders.append(order)
            else:
                orders_to_dispatch.append(order)

        self.orders_scheduled_to_close = []

        for order in orders_to_dispatch:
            self.dispatch(Order(order))

    def plot_balances(self, base, quote, price):
        balance = self.balance.copy()
        exchange_date = datetime.fromtimestamp(self.seconds()).isoformat()
        base_value = balance[base] * price
        equity = balance[quote] + base_value
        ttip = "<b>Total: %.2f %s</b><br />Quote: %.2f %s<br />Base: %.2f %s" % (
            equity, quote,
            balance[quote], quote,
            balance[base], base
        )
        self.dispatch(Plot({
            'name': 'Equity',
            'type': 'scatter',
            'mode': 'lines',
            'color': 'blue',
            'yaxis': 'balance',
            'data': {
                'x': exchange_date,
                'y': equity,
                'tooltip': ttip
            },
        }))

    def funds_enough(self, order):
        base, quote = order['symbol'].split('/')
        if order['type'] == 'limit':
            cost = order['amount'] * order['price']
        else:
            book_side = 'bids' if order['side'] == 'sell' else 'asks'
            cost = order['amount'] * self.orderbooks[order['symbol']][book_side][0][0]

        sum_same_side_orders = sum([o['amount'] for o in self.open_orders.values() if o['side'] == order['side']])

        if order['side'] == 'buy':
            return cost + sum_same_side_orders <= self.balance[quote]
        else:
            return order['amount'] + sum_same_side_orders <= self.balance[base]

    def fetch_ticker(self, symbol, **kwargs):
        if symbol in self.orderbooks:
            if not self.orderbooks[symbol]:
                raise KeyError("TestEX is not ready yet")
            return {"bid": self.orderbooks[symbol]['bids'][0][0], "ask": self.orderbooks[symbol]['asks'][0][0], "timestamp": self.milliseconds()}
        raise NotImplementedError("Symbol %s not found in testex.orderbooks" % symbol)

    def fetch_balance(self):
        balances = {}
        free_balance = self.fetch_free_balance()
        balances['free'] = free_balance
        used_balance = self.fetch_used_balance()
        for asset in used_balance.keys():
            balances[asset] = {'free': free_balance[asset], 'used': used_balance[asset], 'total': self.balance[asset]}
        balances['used'] = used_balance
        balances['total'] = self.balance.copy()
        return balances

    def fetch_free_balance(self):
        free_balance = self.balance.copy()
        for key in self.open_orders.keys():
            order = self.open_orders[key]
            side = order['side']
            base, quote = order['symbol'].split('/')
            if side == "buy":
                free_balance[quote] -= order['amount'] * order['price']
            else:
                free_balance[base] -= order['amount']
        return free_balance

    def fetch_used_balance(self):
        free_balance = self.fetch_free_balance()
        used_balance = self.balance.copy()
        for asset in used_balance.keys():
            used_balance[asset] -= free_balance[asset]
        return used_balance

    def fetch_total_balance(self):
        return self.balance.copy()



import os
import pathlib
import pkgutil
from pathlib import Path

import dotenv


class System:

    @staticmethod
    def get_data_dir():
        APPNAME = "br"
        import sys
        from os import path, environ
        if sys.platform == 'darwin':
            from AppKit import NSSearchPathForDirectoriesInDomains
            # http://developer.apple.com/DOCUMENTATION/Cocoa/Reference/Foundation/Miscellaneous/Foundation_Functions/Reference/reference.html#//apple_ref/c/func/NSSearchPathForDirectoriesInDomains
            # NSApplicationSupportDirectory = 14
            # NSUserDomainMask = 1
            # True for expanding the tilde into a fully qualified path
            appdata = path.join(NSSearchPathForDirectoriesInDomains(14, 1, True)[0], APPNAME)
        elif sys.platform == 'win32':
            appdata = path.join(environ['APPDATA'], APPNAME)
        else:
            appdata = path.expanduser(path.join("~", "." + APPNAME))
        return appdata

    @staticmethod
    def get_root_pkg():
        return __package__.split('.')[0]

    @staticmethod
    def read_file(path):
        return pkgutil.get_data(System.get_root_pkg(), path)

    @staticmethod
    def get_cache_dir():
        return System.get_subpath(['cache'])

    @staticmethod
    def get_import_dir():
        return System.get_subpath(['data', 'imported'])

    @staticmethod
    def get_subpath(subpath: list):
        full_path = [System.get_data_dir()] + subpath
        path = os.path.sep.join(full_path)
        if not os.path.exists(path):
            pathlib.Path(path).mkdir(parents=True, exist_ok=True)
        return path

    @staticmethod
    def load_env(env_dir='.', environment='dev'):
        os.environ['TK_ENV'] = environment
        env = Path(env_dir) / ('.env')
        env_local = Path(env_dir) / ('.env.local')
        env_env = Path(env_dir) / ('.env.%s' % environment)
        env_env_local = Path(env_dir) / ('.env.%s.local' % environment)
        # **.local has precedence, and .env.<env> over .env
        dotenv.load_dotenv(dotenv_path=env_env_local)
        dotenv.load_dotenv(dotenv_path=env_local)
        dotenv.load_dotenv(dotenv_path=env_env)
        dotenv.load_dotenv(dotenv_path=env)



from tradingkit.display.plotter import Plotter
from tradingkit.pubsub.core.event import Event


class NonePlotter(Plotter):
    """ This Plotter does nothing """

    def on_event(self, event: Event):
        pass



from __future__ import annotations

import re
from abc import abstractmethod

from tradingkit.pubsub.core.event import Event
from tradingkit.utils.serializable import Serializable


class State(Serializable):
    """
    We define a state object which provides some utility functions for the
    individual states within the state machine.
    """

    @abstractmethod
    def on_event(self, event: Event) -> State:
        """
        Handle events that are delegated to this State.
        """
        pass

    def on_create(self):
        """
        Executes in the constructor.
        """
        pass

    def __repr__(self):
        """
        Leverages the __str__ method to describe the State.
        """
        return self.__str__()

    def __str__(self):
        """
        Returns the name of the State.
        """
        state_filter = lambda attr: not re.match('^__.+__$', attr) and not callable(getattr(self, attr))
        attrs = {attr: getattr(self, attr) for attr in dir(self) if state_filter(attr)}
        return "%s(%s)" % (self.__class__.__name__, attrs)

    def serialize(self) -> str:
        return str(self)

    @staticmethod
    def unserialize(data: str):
        pass


import importlib
import os
import re

from tradingkit.utils.injector import Injector


class ConfigInjector(Injector):

    var_processors = {
        "int": lambda x: int(x),
        "str": lambda x: str(x),
        "float": lambda x: float(x)
    }

    def __init__(self, config):
        self.config = config
        self.cache = {}

    def inject(self, key, type=None):
        if key in self.cache:
            return self.cache[key]
        elif key in self.config:
            instance = self.inject_config(self.config[key])
            if instance is None:
                raise ValueError("Injection %s not defined, please check configs" % key)
            if type and not isinstance(instance, type):
                raise ValueError("Injection %s(%s) not matching with %s" % (key, instance.__class__, str(type)))
            self.cache[key] = instance
            return instance
        else:
            return None

    def inject_config(self, config):
        if isinstance(config, dict):
            if 'class' in config:
                element_class = ConfigInjector.find_class(config)
                if 'arguments' in config:
                    arguments_config = config['arguments']
                    arguments = self.inject_config(arguments_config)
                    if isinstance(arguments, dict):
                        instance = element_class(**arguments)
                    elif isinstance(arguments, list):
                        instance = element_class(*arguments)
                    else:
                        raise TypeError("Invalid arguments type, it must be list or dict")
                else:
                    instance = element_class()

                if 'calls' in config:
                    calls = config['calls']
                    if not isinstance(calls, dict):
                        raise TypeError("calls must be a dict")
                    for method_name in calls:
                        method = getattr(instance, method_name)
                        arguments_config = calls[method_name]
                        arguments = self.inject_config(arguments_config)
                        if isinstance(arguments, dict):
                            method(**arguments)
                        elif isinstance(arguments, list):
                            method(*arguments)
                        else:
                            raise TypeError("Invalid call arguments type, it must be list or dict")
                return instance

            else:
                return {el: self.inject_config(config[el]) for el in config}
        elif isinstance(config, list):
            return [self.inject_config(el) for el in config]
        elif isinstance(config, str):
            if config[0] == '@':
                return self.inject(config[1:])
            else:
                return ConfigInjector.replace_env(config)
        return config

    @staticmethod
    def find_class(spec: dict):
        spec_split = spec['class'].split(".")
        ref_module_str, ref_class_str = ".".join(spec_split[:-1]), spec_split[-1]
        ref_module = importlib.import_module(ref_module_str)
        return getattr(ref_module, ref_class_str)

    @staticmethod
    def replace_env(env_ref: str):
        matches = re.match("%env\(((int|str|float):)?([A-Za-z_][A-Za-z0-9_]*)(:(.*))?\)%", env_ref)
        if matches:
            if matches.group(2):
                processor = ConfigInjector.var_processors[matches.group(2)]
            else:
                processor = ConfigInjector.var_processors['str']

            var = matches.group(3)
            if var in os.environ:
                return processor(os.environ[var])
            elif matches.group(5) or matches.group(5) == '':
                return processor(matches.group(5))
            else:
                raise KeyError("Undefined ENV var '%s'" % var)

        return env_ref



from abc import ABC

from tradingkit.pubsub.core.subscriber import Subscriber
from tradingkit.pubsub.event.plot import Plot


class Plotter(Subscriber, ABC):

    def subscribed_events(self) -> list:
        return [Plot]



import logging
from abc import ABC, abstractmethod
import websocket

from tradingkit.data.feed.feeder import Feeder
from tradingkit.pubsub.core.publisher import Publisher


class WebsocketFeeder(Feeder, Publisher, ABC):

    def __init__(self, symbol, credentials, url):
        super().__init__()
        self.symbol = symbol
        if credentials is not None and ('apiKey' and 'secret') not in credentials:
            raise KeyError("credentials must contain apiKey and secret")
        self.credentials = credentials
        self.url = url
        self.keep_running = True

    @abstractmethod
    def on_open(self, ws):
        pass

    @abstractmethod
    def on_message(self, ws, message):
        pass

    def on_error(self, ws, error):
        logging.info("[Websocket error] %s" % str(error))
        raise error

    def on_close(self, ws):
        logging.info("WebSocket closed")

    def feed(self):
        ws = websocket.WebSocketApp(
            url=self.url,
            on_message=self.on_message,
            on_open=self.on_open,
            on_error=self.on_error,
            on_close=self.on_close,
        )
        ws.run_forever(ping_interval=15, ping_timeout=10)




from abc import abstractmethod


class Serializable:
    @abstractmethod
    def serialize(self) -> str:
        pass

    @staticmethod
    @abstractmethod
    def unserialize(data: str):
        pass


import json
from tradingkit.data.feed.websocket_feeder import WebsocketFeeder
from tradingkit.pubsub.event.book import Book
from tradingkit.pubsub.event.trade import Trade


class PublicKrakenFeeder(WebsocketFeeder):

    # Converts symbols from normal to kraken vocab
    denormalized_symbol = {
        "BTC/EUR": "XBT/EUR",
        "BTC/USD": "XBT/USD",
        "BTC/USDT": "XBT/USDT",
    }

    # Converts symbols from kraken to normal vocab
    normalized_symbol = {
        "XBT/EUR": "BTC/EUR",
        "XBT/USD": "BTC/USD",
        "XBT/USDT": "BTC/USDT",
    }

    def __init__(self, symbol):
        super().__init__(symbol, None, "wss://ws.kraken.com")
        self.orderbooks = {}

    def on_open(self, ws):
        ws.send(json.dumps({
            'event': 'subscribe',
            'subscription': {'name': 'trade'},
            'pair': [self.denormalized_symbol[self.symbol]]
        }))
        ws.send(json.dumps({
            'event': 'subscribe',
            'subscription': {'name': 'book'},
            'pair': [self.denormalized_symbol[self.symbol]]
        }))

    def on_message(self, ws, message):
        data = json.loads(message)
        if "trade" in data:
            symbol = self.normalized_symbol[data[-1]]
            for trade in data[1]:
                price = float(trade[0])
                amount = float(trade[1])
                cost = float(trade[0]) * float(trade[1])
                timestamp = int(float(trade[2]) * 1000)
                side = 'buy' if trade[3] == 'b' else 'sell'
                type = 'market' if trade[4] == 'm' else 'limit'
                trade_data = {
                    'price': price,
                    'amount': amount,
                    'cost': cost,
                    'timestamp': timestamp,
                    'side': side,
                    'type': type,
                    'symbol': symbol
                }
                self.dispatch(Trade(trade_data))

        elif "book-10" in data:
            keys = data[1].keys()
            symbol = self.normalized_symbol[data[-1]]
            if "as" in keys:
                self.orderbooks[symbol] = {
                    "bids": [
                        [
                            float(data[1]["bs"][0][0]),
                            float(data[1]["bs"][0][1])
                        ]
                    ],
                    "asks": [
                        [
                            float(data[1]["as"][0][0]),
                            float(data[1]["as"][0][1])
                        ]
                    ],
                    "timestamp": int(float(data[1]["as"][0][2]) * 1000),
                    "symbol": symbol
                }
            else:
                if "a" in keys:
                    self.orderbooks[symbol]["asks"] = [
                        [
                            float(data[1]["a"][0][0]),
                            float(data[1]["a"][0][1])
                        ]
                    ]
                    self.orderbooks[symbol]["timestamp"] = int(float(data[1]["a"][0][2]) * 1000)
                    self.orderbooks[symbol]["symbol"] = symbol
                if "b" in keys:
                    self.orderbooks[symbol]["bids"] = [
                        [
                            float(data[1]["b"][0][0]),
                            float(data[1]["b"][0][1])
                        ]
                    ]
                    self.orderbooks[symbol]["timestamp"] = int(float(data[1]["b"][0][2]) * 1000)
                    self.orderbooks[symbol]["symbol"] = symbol
            self.dispatch(Book(self.orderbooks[symbol]))



from tradingkit.data.feed.feeder import Feeder
from tradingkit.pubsub.core.publisher import Publisher
from tradingkit.pubsub.event.trade import Trade


class ListFeeder(Feeder, Publisher):
    def __init__(self, trades):
        super().__init__()
        self.trades = trades

    def feed(self):
        for trade in self.trades:
            self.dispatch(Trade(trade))



import importlib.resources as res
import json
import logging
import os
import sys
from datetime import datetime

import ccxt
import docopt
import pkg_resources
from ccxt import Exchange
from dateutil.relativedelta import relativedelta
from setuptools import Command

from tradingkit.cli.runner import Runner
from tradingkit.cli.optimizer import Optimizer
from tradingkit.data.feed.feeder import Feeder
from tradingkit.display.plotter import Plotter
from tradingkit.strategy.strategy import Strategy
from tradingkit.utils.config_injector import ConfigInjector
from tradingkit.utils.system import System
from tradingkit.data.fetch.ccxt_fetcher import CCXTFetcher


class CLI(Command):
    """Trading Kit CLI.

    Usage:
      tk run [<strategy_dir>] [-e <env>] [-y <year>] [-m <month>] [--optimize | --plot | --live_plot] [ --show-open-orders | --show-orders] [--loglevel <level>]
      tk import [-x <exchange>] [-s <symbol>] [-y <year>] [-m <month>] [--no-funding | --only-funding]
      tk --help
      tk --version

    Options:
      -e, --env <env>                    Select environment [default: dev].
      -x, --exchange <exchange>          Select exchange.
      -s, --symbol <symbol>              Select the exchange symbol.
      -y, --year <year>                  Year to backtest.
      -m, --month <month>                Month to backtest.
      -o, --optimize                     Run the optimizer for the strategy.
      --plot                             Plot the results.
      --live_plot                        Plot real-time charts.
      --show-open-orders                 Show closed and open orders on chart.
      --show-orders                      Show only closed orders on chart.
      --no-funding                       Import only trades data
      --only-funding                     Import only funding rate data
      --loglevel <level>                 Sets the log level [default: error].
      -h, --help                         Show this screen.
      -v, --version                      Show version.

    """

    @staticmethod
    def main():
        sys.path.append(os.getcwd())
        version = str(pkg_resources.require("tradingkit")[0])
        args = docopt.docopt(str(CLI.__doc__), version=version)

        logging.basicConfig(level=logging.getLevelName(args['--loglevel'].upper()))

        if args['import']:
            symbol = args['--symbol']
            if not args['--no-funding']:
                exchange_name = 'bitmex'
                exchange_class = getattr(ccxt, exchange_name)
                exchange = exchange_class()
                CLI.command_import_funding_rate(exchange_name, exchange, symbol)
            if not args['--only-funding']:
                exchange_name = args['--exchange']
                exchange_class = getattr(ccxt, exchange_name)
                exchange = exchange_class()
                fetcher = CCXTFetcher(exchange)
                year = int(args['--year'])
                months = [int(args['--month'])] if args['--month'] else range(1, 13)
                CLI.command_import(exchange_name, symbol, fetcher, year, months)
        elif args['--optimize']:
            Optimizer().optimize(args)
        else:

            config = json.loads(res.read_text("tradingkit.config", "config.json"))

            strategy_dir = args['<strategy_dir>'] or '.'
            System.load_env(strategy_dir, args['--env'])

            main_conf = "%s/config/config.json" % os.path.abspath(strategy_dir)
            if os.path.exists(main_conf):
                config.update(json.load(open(main_conf, 'r')))

            env_conf = "%s/config/config.%s.json" % (os.path.abspath(strategy_dir), args['--env'])
            if os.path.exists(env_conf):
                config.update(json.load(open(env_conf, 'r')))

            if args['--year']:
                if args['--month']:
                    since = datetime.fromisoformat(
                        "%d-%02d-01T00:00:00+00:00" % (int(args['--year']), int(args['--month']))
                    )
                    to = since + relativedelta(months=1)
                else:
                    since = datetime.fromisoformat("%d-01-01T00:00:00+00:00" % int(args['--year']))
                    to = since + relativedelta(years=1)
                config['since'] = since.isoformat()
                config['to'] = to.isoformat()

            injector = ConfigInjector(config)
            feeder = injector.inject('feeder', Feeder)
            exchange = injector.inject('exchange', Exchange)
            plotter = injector.inject('plotter', Plotter)
            if args['--live_plot']:
                plotter.set_live()
            if args['--show-open-orders']:
                plotter.set_chart_type(2)
            elif args['--show-orders']:
                plotter.set_chart_type(1)
            strategy = injector.inject('strategy', Strategy)
            bridge = injector.inject('bridge', Exchange)
            feeder_adapters = injector.inject('feeder_adapters', list)

            Runner.run(feeder, exchange, plotter, strategy, bridge, feeder_adapters, args['--optimize'], args['--plot'])

    @staticmethod
    def command_import(exchange_name, symbol, fetcher, year, months):
        for month in months:
            next_period_year = year if month < 12 else (year + 1)
            next_period_month = (month + 1) if month < 12 else 1
            trades, candles = fetcher.fetch_all(
                symbol=symbol,
                since8601="%d-%02d-01T00:00:00+00:00" % (year, month),
                to8601="%d-%02d-01T00:00:00+00:00" % (next_period_year, next_period_month)
            )
            import_dir = System.get_import_dir()
            base, quote = symbol.split('/')
            full_filename = '%s/%s-%s_%s-%d-%02d.json' % (import_dir, exchange_name, base, quote, year, month)
            json.dump(trades, open(full_filename, 'w'))

            for timeframe in candles.keys():
                full_filename = '%s/%s-%s_%s-%d-%02d_%s_candle.json' % (import_dir, exchange_name, base, quote, year, month, timeframe)
                json.dump(candles[timeframe], open(full_filename, 'w'))

    @staticmethod
    def command_import_funding_rate(exchange_name, exchange, symbol):
        pairs_conversion = {'BTC/USD': 'XBTUSD'}
        if symbol in pairs_conversion:
            funding_data = exchange.public_get_funding({'symbol': pairs_conversion[symbol], 'count': 500})
            resp = exchange.public_get_funding({'symbol': pairs_conversion[symbol], 'count': 500, 'start': len(funding_data)})
            while len(resp) > 0:
                for rate in resp:
                    funding_data.append(rate)
                resp = exchange.public_get_funding({'symbol': pairs_conversion[symbol], 'count': 500, 'start': len(funding_data)})

            import_dir = System.get_import_dir()
            base, quote = symbol.split('/')
            full_filename = '%s/%s_%s-%s_%s.json' % (import_dir, 'funding', exchange_name, base, quote)
            json.dump(funding_data, open(full_filename, 'w'))



import base64
import hashlib
import hmac
import json
import sys
import os
import threading
import time
import urllib.request
from datetime import datetime

from websocket import create_connection, _logging

from tradingkit.data.feed.feeder import Feeder
from tradingkit.pubsub.core.publisher import Publisher
from tradingkit.pubsub.event.book import Book
from tradingkit.pubsub.event.candle import Candle
from tradingkit.pubsub.event.order import Order
from tradingkit.pubsub.event.trade import Trade


class KrakenFeeder(Feeder, Publisher):

    def __init__(self, credentials=None, ignore_outdated=False, pair=None):
        super().__init__()
        if pair is None:
            pair = {'symbol': 'BTC/EUR'}
        self.public_ws = None
        self.private_ws = None
        if credentials is not None:
            if ('apiKey' and 'secret') not in credentials:
                raise KeyError("credentials must contain apiKey and secret")
        self.credentials = credentials

        # Converts symbols from normal to kraken vocab
        self.denormalized_symbol = {
            "BTC/EUR": "XBT/EUR",
            "BTC/USD": "XBT/USD",
            "BTC/USDT": "XBT/USDT",
        }

        # Converts symbols from kraken to normal vocab
        self.normalized_symbol = {
            "XBT/EUR": "BTC/EUR",
            "XBT/USD": "BTC/USD",
            "XBT/USDT": "BTC/USDT",
        }
        self.symbol = self.denormalized_symbol[pair['symbol']]  # used to send requests to kraken
        self.on_open()
        self.ignore_outdated = ignore_outdated
        self.orderbooks = {}
        self.lock = None
        self.candle = {'id': '', 'data': {}}

    def authenticate(self):
        api_nonce = bytes(str(int(time.time() * 1000)), "utf-8")
        api_request = urllib.request.Request("https://api.kraken.com/0/private/GetWebSocketsToken",
                                             b"nonce=%s" % api_nonce)
        api_request.add_header("API-Key", self.credentials['apiKey'])
        api_request.add_header("API-Sign", base64.b64encode(hmac.new(base64.b64decode(self.credentials['secret']),
                                                                     b"/0/private/GetWebSocketsToken" + hashlib.sha256(
                                                                         api_nonce + b"nonce=%s" % api_nonce).digest(),
                                                                     hashlib.sha512).digest()))
        resp = json.loads(urllib.request.urlopen(api_request).read())
        if 'result' in resp and 'token' in resp['result']:
            resp = resp['result']['token']
        return resp

    def on_open(self):
        api_domain = "wss://ws.kraken.com/"
        auth_api_domain = "wss://ws-auth.kraken.com"

        try:
            self.public_ws = create_connection(api_domain)
        except Exception as error:
            _logging.warning("WebSocket connection failed (%s)" % error)
            time.sleep(600)
            self.on_open()
        try:
            self.private_ws = create_connection(auth_api_domain)
        except Exception as error:
            _logging.warning("WebSocket connection failed (%s)" % error)
            time.sleep(600)
            self.on_open()
        token = self.authenticate()
        self.subscribe(token)

    def subscribe(self, token):
        api_feed = "book"
        api_depth = 10
        book_feed = '{"event":"subscribe", "subscription":{"name":"%(feed)s", "depth":%(depth)s}, "pair":["%(symbol)s"]}' % {
            "feed": api_feed, "depth": api_depth, "symbol": self.symbol}
        trade_feed = '{"event": "subscribe", "pair": ["%(symbol)s"],  "subscription": {"name": "trade", "token": "%(token)s"}}' % {
            "symbol": self.symbol, 'token': token}
        own_trades_feed = '{"event": "subscribe", "subscription": {"name": "ownTrades","token": "%(token)s"}}' % {
            'token': token}

        try:
            self.public_ws.send(trade_feed)
            self.public_ws.send(book_feed)
            self.private_ws.send(own_trades_feed)
        except Exception as error:
            _logging.warning("Feed subscription failed (%s)" % error)
            self.public_ws.close()
            self.private_ws.close()
            sys.exit(1)

    def dispatch_event(self, event):
        self.lock.acquire()
        self.dispatch(event)
        self.lock.release()

    def on_message(self, message):
        ts = time.time()
        if "ownTrades" in message:
            for dict in message[0]:
                for order in dict:
                    if ts - float(dict[order]['time']) < 10:  # filter orders since 10 seg
                        order_data = {
                            'id': dict[order]['ordertxid'],
                            'timestamp': int(float(dict[order]['time']) * 1000),
                            'lastTradeTimestamp': int(float(dict[order]['time']) * 1000),
                            'status': 'filled',
                            'symbol': self.normalized_symbol[dict[order]['pair']],
                            'type': dict[order]['ordertype'],
                            'side': dict[order]['type'],
                            'price': float(dict[order]['price']),
                            'amount': float(dict[order]['vol'])
                        }
                        self.dispatch_event(Order(order_data))
        elif "book-10" in message:
            keys = message[1].keys()
            symbol = self.normalized_symbol[message[-1]]
            if "as" in keys:
                self.orderbooks[symbol] = {
                    "bids": [
                        [
                            float(message[1]["bs"][0][0]),
                            float(message[1]["bs"][0][1])
                        ]
                    ],
                    "asks": [
                        [
                            float(message[1]["as"][0][0]),
                            float(message[1]["as"][0][1])
                        ]
                    ],
                    "timestamp": int(float(message[1]["as"][0][2]) * 1000),
                    "symbol": symbol
                }
            else:
                if "a" in keys:
                    self.orderbooks[symbol]["asks"] = [
                        [
                            float(message[1]["a"][0][0]),
                            float(message[1]["a"][0][1])
                        ]
                    ]
                    self.orderbooks[symbol]["timestamp"] = int(float(message[1]["a"][0][2]) * 1000)
                    self.orderbooks[symbol]["symbol"] = symbol
                if "b" in keys:
                    self.orderbooks[symbol]["bids"] = [
                        [
                            float(message[1]["b"][0][0]),
                            float(message[1]["b"][0][1])
                        ]
                    ]
                    self.orderbooks[symbol]["timestamp"] = int(float(message[1]["b"][0][2]) * 1000)
                    self.orderbooks[symbol]["symbol"] = symbol
            self.dispatch_event(Book(self.orderbooks[symbol]))
        elif "trade" in message:
            symbol = self.normalized_symbol[message[-1]]
            for trade in message[1]:
                price = float(trade[0])
                amount = float(trade[1])
                cost = float(trade[0]) * float(trade[1])
                timestamp = int(float(trade[2]) * 1000)
                side = 'buy' if trade[3] == 'b' else 'sell'
                type = 'market' if trade[4] == 'm' else 'limit'
                trade_data = {
                    'price': price,
                    'amount': amount,
                    'cost': cost,
                    'timestamp': timestamp,
                    'side': side,
                    'type': type,
                    'symbol': symbol
                }
                self.dispatch_event(Trade(trade_data))

    def run(self, is_private):
        if is_private:
            _ws = self.private_ws
        else:
            _ws = self.public_ws
        while True:
            ws_data = "No Data."
            try:
                ws_data = _ws.recv()
                if ws_data:
                    message = json.loads(ws_data)
                    self.on_message(message)
            except KeyboardInterrupt:
                _ws.close()
                sys.exit(0)
            except Exception as error:
                _logging.warning("[WebSocket error] %s" % str(error))
                _logging.warning("[WebSocket data] %s" % str(ws_data))
                time.sleep(60)
                self.on_open()
                if is_private:
                    _ws = self.private_ws
                else:
                    _ws = self.public_ws

    def feed(self):
        # creating a lock
        self.lock = threading.Lock()

        # creating threads
        public_t = threading.Thread(target=self.run, args=(False,))
        private_t = threading.Thread(target=self.run, args=(True,))

        # start threads
        public_t.start()
        private_t.start()

        # wait until threads finish their job
        public_t.join()
        _logging.warning("[WebSocket data public STOP] %s" % str(public_t))
        private_t.join()
        _logging.warning("[WebSocket data private STOP] %s" % str(private_t))





import os
import re
import socketserver
from http.server import SimpleHTTPRequestHandler
from typing import Tuple, Optional, Union


class RequestHandler(SimpleHTTPRequestHandler):

    def __init__(self, routing: dict, *args, **kwargs):
        self.routing = routing if routing else None
        super().__init__(*args, **kwargs)

    def translate_path(self, path):
        if self.routing:
            for pattern in self.routing:
                if path.startswith(pattern):
                    return os.path.join(self.routing[pattern], path[len(pattern):])
        return super().translate_path(path)



from tradingkit.pubsub.core.event import Event
from tradingkit.pubsub.core.subscriber import Subscriber


class Publisher:

    def __init__(self):
        self.events = {}

    def register(self, subscriber: Subscriber):
        for event in subscriber.subscribed_events():
            if event not in self.events:
                self.events[event] = []
            self.events[event].append(subscriber)

    def unregister(self, subscriber: Subscriber):
        for event in subscriber.subscribed_events():
            self.events[event].remove(subscriber)

    def dispatch(self, event: Event):
        if event.__class__ in self.events:
            for subscriber in self.events[event.__class__]:
                subscriber.on_event(event)



from tradingkit.pubsub.core.event import Event
from tradingkit.strategy.strategy import Strategy


class StrategyEvent(Event):
    def __init__(self, payload, strategy: Strategy):
        super().__init__(payload)
        self.strategy = strategy



import math
import time
from datetime import datetime
import json
import os
import numpy as np
from ccxt import Exchange

from tradingkit.cli.runner import Runner
from tradingkit.data.feed.feeder import Feeder
from tradingkit.display.plotter import Plotter
from tradingkit.strategy.strategy import Strategy
from tradingkit.utils.config_injector import ConfigInjector
from tradingkit.utils.system import System


class Optimizer:

    def __init__(self):
        self.args = None
        self.param_names = []
        self.population_size = 10
        self.count = 0
        self.max_iterations = 100
        self.max_iteration_without_improv = 10
        self.start_time = time.time()

    def objective_function(self, genome):
        self.count += 1
        total_profit = 0
        results = []
        # select the dataset/s to optimize

        #2016
        since = datetime.fromisoformat("2016-01-01 00:00:00+00:00")
        to = datetime.fromisoformat("2017-01-01 00:00:00+00:00")
        result = self.run_simulation(since.isoformat(), to.isoformat(), genome)

        if result['start_base_balance'] == 0:
            profit = (result['end_equity'] - result['start_equity']) / result['start_equity'] * 100
        else:
            profit = (result['base_balance'] - result['start_base_balance']) / result['start_base_balance'] * 100
        if result['end_equity'] > 0:

            results.append(profit)
        else:
            return np.random.uniform(-200, -100, 1)[0]
        total_profit += profit

        # 2017
        since = datetime.fromisoformat("2017-01-01 00:00:00+00:00")
        to = datetime.fromisoformat("2018-01-01 00:00:00+00:00")
        result = self.run_simulation(since.isoformat(), to.isoformat(), genome)

        if result['start_base_balance'] == 0:
            profit = (result['end_equity'] - result['start_equity']) / result['start_equity'] * 100
        else:
            profit = (result['base_balance'] - result['start_base_balance']) / result['start_base_balance'] * 100
        if result['end_equity'] > 0:

            results.append(profit)
        else:
            return np.random.uniform(-200, -100, 1)[0]
        total_profit += profit

        # 2018
        since = datetime.fromisoformat("2018-01-01 00:00:00+00:00")
        to = datetime.fromisoformat("2019-01-01 00:00:00+00:00")
        result = self.run_simulation(since.isoformat(), to.isoformat(), genome)

        if result['start_base_balance'] == 0:
            profit = (result['end_equity'] - result['start_equity']) / result['start_equity'] * 100
        else:
            profit = (result['base_balance'] - result['start_base_balance']) / result['start_base_balance'] * 100
        if result['end_equity'] > 0:

            results.append(profit)
        else:
            return np.random.uniform(-200, -100, 1)[0]
        total_profit += profit

        # 2019
        since = datetime.fromisoformat("2019-01-01 00:00:00+00:00")
        to = datetime.fromisoformat("2020-01-01 00:00:00+00:00")
        result = self.run_simulation(since.isoformat(), to.isoformat(), genome)

        if result['start_base_balance'] == 0:
            profit = (result['end_equity'] - result['start_equity']) / result['start_equity'] * 100
        else:
            profit = (result['base_balance'] - result['start_base_balance']) / result['start_base_balance'] * 100
        if result['end_equity'] > 0:

            results.append(profit)
        else:
            return np.random.uniform(-200, -100, 1)[0]
        total_profit += profit


        median_result = np.median(results)
        min_result = min(results)
        result = (median_result + min_result * 7) / 8

        if result > 0:
            c_handle = open(str(self.start_time) + "_out.csv", 'a')
            data = []
            data.append(self.count)
            data.append("         T prof:")
            data.append(total_profit)
            data.append("%          ")
            data.append(genome.tolist())
            data.append("          ")
            data.append(datetime.now())
            data.append("   ")
            data.append(result)
            data.append(results)
            np.savetxt(c_handle, [data], delimiter="  ", fmt="%s")
            c_handle.close()
        return result

    def run_simulation(self, since, to, genome):
        config = json.loads(System.read_file("config/config.json"))
        strategy_dir = self.args['<strategy_dir>'] or '.'
        System.load_env(strategy_dir, self.args['--env'])

        main_conf = "%s/config/config.json" % os.path.abspath(strategy_dir)
        if os.path.exists(main_conf):
            config.update(json.load(open(main_conf, 'r')))

        env_conf = "%s/config/config.%s.json" % (os.path.abspath(strategy_dir), self.args['--env'])
        if os.path.exists(env_conf):
            config.update(json.load(open(env_conf, 'r')))

        config['since'] = since
        config['to'] = to
        for i in range(len(self.param_names)):
            config['config'][self.param_names[i]] = genome[i]

        injector = ConfigInjector(config)
        feeder = injector.inject('feeder', Feeder)
        exchange = injector.inject('exchange', Exchange)
        plotter = injector.inject('plotter', Plotter)

        strategy = injector.inject('strategy', Strategy)
        bridge = injector.inject('bridge', Exchange)
        feeder_adapters = injector.inject('feeder_adapters', list)

        result = Runner.run(feeder, exchange, plotter, strategy, bridge, feeder_adapters, self.args['--optimize'],
                            self.args['--plot'])
        return result

    def optimize(self, args):
        strategy_dir = args['<strategy_dir>'] or '.'
        route = "%s/config/config.json" % os.path.abspath(strategy_dir)
        if os.path.exists(route):
            config = json.load(open(route, 'r'))
        else:
            raise ValueError("The path: %s due not exist." % str(route))
        varbound = []
        vartype = []
        for param in config['optimizer_config']:
            self.param_names.append(param)
            varbound.append([config['optimizer_config'][param]['from'], config['optimizer_config'][param]['to']])
            vartype.append([config['optimizer_config'][param]['type']])
        self.args = args

        t = datetime.now()

        # first evaluation
        population = self.generate_population(varbound)
        score = {}

        for genome in population:
            genome = np.array(genome)

            _t = datetime.now()
            profit = self.objective_function(genome)
            print("Iteration Time: ", datetime.now() - _t)
            score[profit] = genome

        print("Lapsed Time: ", datetime.now() - t)

        print("Top 10 solutions:")
        best_sol = 0
        i = 0
        for key, value in sorted(score.items(), reverse=True):
            if i < 10:
                if i == 0:
                    best_sol = key
                print(i, key, value)
                i += 1

        # next n evaluations
        iteration_without_improv = 0
        for iteration in range(self.max_iterations):
            if iteration_without_improv > self.max_iteration_without_improv:
                break
            score = self.mutate_and_evaluate(varbound, score, iteration)

            print("Lapsed Time: ", datetime.now() - t)

            print(iteration, "Top 10 solutions:")
            f_handle = open(str(self.start_time) + "_best.csv", 'a')
            np.savetxt(f_handle, [[iteration]], delimiter="  ", fmt="%s")
            f_handle.close()
            i = 0
            for key, value in sorted(score.items(), reverse=True):
                if i < 10:
                    if i == 0:
                        if best_sol < key:
                            best_sol = key
                            iteration_without_improv = 0
                        else:
                            iteration_without_improv += 1
                    print(i, key, value)
                    f_handle = open(str(self.start_time) + "_best.csv", 'a')
                    np.savetxt(f_handle, [[key, value]], delimiter="  ", fmt="%s")
                    f_handle.close()
                    i += 1
        print("Start time:", t)
        print("End time:", datetime.now())

    def generate_population(self, varbound):
        population = []
        for i in range(self.population_size):
            genome = []
            for bound in varbound:
                genome.append(np.random.uniform(bound[0], bound[1], 1)[0])
            population.append(genome)
        return population

    def mutate_and_evaluate(self, varbound, score, iteration):
        new_score = {}
        i = 0
        for key, value in sorted(score.items(), reverse=True):
            if i < self.population_size:
                new_score[key] = value
                new_genome = value.tolist()
                index = np.random.randint(len(new_genome), size=1)[0]
                ratio = max(iteration - 10, 0)  # first 10 iterations use max mutation speed and decrease after
                new_genome[index] = (new_genome[index] * ratio +
                                     np.random.uniform(varbound[index][0], varbound[index][1], 1)[0]) / (ratio + 1)
                new_genome = np.array(new_genome)

                _t = datetime.now()
                profit = self.objective_function(new_genome)
                print("Iteration Time: ", datetime.now() - _t)

                new_score[profit] = new_genome
                i += 1
            else:
                return new_score

        return new_score





from abc import abstractmethod

from tradingkit.pubsub.core.event import Event


class Subscriber:

    @abstractmethod
    def subscribed_events(self) -> list:
        pass

    @abstractmethod
    def on_event(self, event: Event):
        pass



from logging import Handler, Formatter
import datetime
from tradingkit.notifications.telegram_api import TelegramAPI


class RequestsHandler(Handler):

    def __init__(self, bot_token, bot_chat_id):
        self.bot = TelegramAPI(bot_token, bot_chat_id)
        super().__init__()

    def emit(self, record):
        if record.levelname == 'WARNING':
            log_entry = self.format(record)
            self.bot.send_text(log_entry)


class LogstashFormatter(Formatter):
    def __init__(self):
        super(LogstashFormatter, self).__init__()

    def format(self, record):
        t = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')

        return "INFO: {message}".format(message=record.msg, datetime=t)



import logging
import os
import shutil
import random
from datetime import datetime
import json

from tradingkit.display.highstock.server.plot_server import PlotServer
from tradingkit.display.plotter import Plotter
from tradingkit.pubsub.core.event import Event
from tradingkit.pubsub.event.plot import Plot
from tradingkit.utils.system import System
from tradingkit.utils.web_server import WebServer

import asyncio
import websockets
import time
import multiprocessing


class HighstockPlotter(Plotter):

    def __init__(self, live_plot=False):
        self.price = []
        self.orders = []
        self.balance = {"quote": 0,
                        "base": 0,
                        "position_vol": 0,
                        "quote_balance": 0,
                        "base_balance": 0,
                        "position_price": 0,
                        "invested": 0}
        self.series = {
            'OHLC': {'name': 'OHLC', 'type': 'candlestick', 'tooltip': 'OHLC', 'data': []},
            'volume': {'name': 'Volume', 'type': 'line', 'tooltip': '', 'data': []},
            'position': {'name': 'Position', 'type': 'line', 'tooltip': '', 'data': []},
            'position_price': {'name': 'Position Price', 'type': 'area', 'tooltip': '', 'data': []},
            'liq_price': {'name': 'Liquidation Price', 'type': 'line', 'tooltip': '', 'data': []},
            'equity': {'name': 'Equity', 'type': 'area', 'tooltip': '', 'data': []},
            'base_equity': {'name': 'Base Equity', 'type': 'area', 'tooltip': '', 'data': []},
            'quote_balance': {'name': 'Quote Balance', 'type': 'line', 'tooltip': '', 'data': []},
            'base_balance': {'name': 'Base Balance', 'type': 'line', 'tooltip': '', 'data': []},
            'hold': {'name': 'Hold', 'type': 'area', 'tooltip': '', 'data': []},
            'buy': {'name': 'Buy', 'type': 'scatter', 'tooltip': '', 'data': []},
            'open_buy': {'name': 'Buy', 'type': 'scatter', 'tooltip': '', 'data': []},
            'cancel_buy': {'name': 'Buy', 'type': 'scatter', 'tooltip': '', 'data': []},
            'sell': {'name': 'Sell', 'type': 'scatter', 'tooltip': '', 'data': []},
            'open_sell': {'name': 'Sell', 'type': 'scatter', 'tooltip': '', 'data': []},
            'cancel_sell': {'name': 'Sell', 'type': 'scatter', 'tooltip': '', 'data': []},
            'buy_series': {'name': 'OpenBuy', 'type': 'line', 'tooltip': '', 'data': [], 'series': [[]]},
            'sell_series': {'name': 'OpenSell', 'type': 'line', 'tooltip': '', 'data': [], 'series': [[]]},
            'assets': None
        }
        self.live_plot = live_plot
        if live_plot:
            self.set_live()
        self.initial_base_balance = 0
        self.chart_port = 6789
        self.balance_port = 6788
        self.chart_type = 0

    def set_chart_type(self, type):
        self.chart_type = type

    def set_live(self):
        self.live_plot = True
        chart_ps = PlotServer(self.chart_port)
        balance_ps = PlotServer(self.balance_port)
        chart_process = multiprocessing.Process(target=chart_ps.run)
        balance_process = multiprocessing.Process(target=balance_ps.run)
        chart_process.start()
        logging.info("Chart PlotServer running at port %s ... \n" % str(self.chart_port))
        balance_process.start()
        logging.info("Balance PlotServer running at port %s ... \n" % str(self.balance_port))

    def on_event(self, event: Event):
        if isinstance(event, Plot):
            plot = event.payload
            if plot['name'].lower() == 'price':
                if plot['data']['timeframe'] == '1m':
                    date = int(datetime.timestamp(datetime.fromisoformat(plot['data']['datetime']))) * 1000
                    self.series['OHLC']['data'].append([date,
                                                        plot['data']['open'],
                                                        plot['data']['high'],
                                                        plot['data']['low'],
                                                        plot['data']['close']])
                    if plot['data']['liquidationPrice'] is not None and 0.8 < float(plot['data']['liquidationPrice']) / plot['data']['close'] < 1.2:
                        liq_price = float(plot['data']['liquidationPrice'])
                    else:
                        liq_price = None
                    self.series['liq_price']['data'].append([date, liq_price])
                    self.series['volume']['data'].append([date, plot['data']['vol']])
                    y = round(self.balance['quote_balance'] + self.balance['base_balance'] * plot['data']['close'])
                    self.series['equity']['data'].append([date, y])

                    self.series['base_equity']['data'].append([date,
                        self.balance['base_balance'] + self.balance['quote_balance'] / plot['data']['close']])
                    self.series['position']['data'].append([date, self.balance['position_vol'] + 0])
                    self.series['position_price']['data'].append([date, self.balance['position_price']])
                    self.series['quote_balance']['data'].append([date, round(self.balance['quote_balance'], 2)])
                    self.series['base_balance']['data'].append([date, self.balance['base_balance']])
                    self.series['hold']['data'].append([date, round(self.series['base_equity']['data'][0][1] *
                        plot['data']['close']) if self.series['base_equity']['data'] else round(
                        plot['data']['base_equity'] * plot['data']['close'])])

                    if self.live_plot:
                        time.sleep(1)
                        self.new_price()
            elif plot['name'].lower() == 'equity':
                if self.series['assets'] is None:
                    self.series['assets'] = {'quote': plot['quote'], 'base': plot['base']}
                    if self.live_plot:
                        asyncio.get_event_loop().run_until_complete(self.update_plot({"action": "produce",
                                                                                      "type": "assets",
                                                                                      "data": self.series['assets']
                                                                                      }, self.chart_port))
                        asyncio.get_event_loop().run_until_complete(self.update_plot({"action": "produce",
                                                                                      "type": "assets",
                                                                                      "data": self.series['assets']
                                                                                      }, self.balance_port))
                self.balance['quote'] = round(plot['data']['y'], 2)
                self.balance['base'] = plot['data']['base_equity']
                self.balance['quote_balance'] = plot['data']['quote_balance']
                self.balance['base_balance'] = plot['data']['base_balance']
                self.balance['position_vol'] = plot['data']['position_vol']
                self.balance['position_price'] = plot['data']['position_price'] if plot['data'][
                                                                       'position_price'] != 0 else None
                self.balance['invested'] = plot['data']['invested']

            elif plot['name'].lower() == 'buy':
                if self.chart_type > 0:
                    order_data = self.format_order_data(plot)
                    self.series['buy']['data'].append(order_data)
                    if self.chart_type > 1:
                        self.close_line(order_data, self.series['buy_series']['series'])

                    if self.live_plot:
                        asyncio.get_event_loop().run_until_complete(self.update_plot({"action": "produce",
                                                                                      "type": 'buy',
                                                                                      "data": order_data
                                                                                      }, self.chart_port))
            elif plot['name'].lower() == 'sell':
                if self.chart_type > 0:
                    order_data = self.format_order_data(plot)
                    self.series['sell']['data'].append(order_data)
                    if self.chart_type > 1:
                        self.close_line(order_data, self.series['sell_series']['series'])
                    if self.live_plot:
                        asyncio.get_event_loop().run_until_complete(self.update_plot({"action": "produce",
                                                                                      "type": 'sell',
                                                                                      "data": order_data
                                                                                      }, self.chart_port))
            elif plot['name'].lower() == 'open_buy':
                if self.chart_type > 1:
                    order_data = self.format_order_data(plot)
                    self.series['open_buy']['data'].append(order_data)
                    self.open_line(order_data, self.series['buy_series']['series'])

            elif plot['name'].lower() == 'open_sell':
                if self.chart_type > 1:
                    order_data = self.format_order_data(plot)
                    self.series['open_sell']['data'].append(order_data)
                    self.open_line(order_data, self.series['sell_series']['series'])

            elif plot['name'].lower() == 'cancel_buy':
                if self.chart_type > 1:
                    order_data = self.format_order_data(plot)
                    self.series['cancel_buy']['data'].append(order_data)

                    self.close_line(order_data, self.series['buy_series']['series'])

            elif plot['name'].lower() == 'cancel_sell':
                if self.chart_type > 1:
                    order_data = self.format_order_data(plot)
                    self.series['cancel_sell']['data'].append(order_data)

                    self.close_line(order_data, self.series['sell_series']['series'])

            else:
                if plot['name'] not in self.series:
                    self.series[plot['name']] = plot.copy()
                    self.series[plot['name']]['data'] = []
                data = plot['data'].copy()
                if 'tooltip' not in data:
                    data['tooltip'] = data['y']

    def format_order_data(self, plot):
        date = int(datetime.timestamp(datetime.fromisoformat(plot['data']['x']))) * 1000
        order_data = [date, plot['data']['y'], plot['data']['type'] + ": " + str(plot['data']['tooltip']),
                      plot['data']['id']]
        return order_data

    def open_line(self, order_data, series):
        for serie in series:
            if not serie or not serie[-1][1]:
                serie.append(order_data)
                return True
        series.append([order_data])

    def close_line(self, order_data, series):
        for serie in series:
            if serie and serie[-1][3] == order_data[3]:
                serie.append(order_data)
                serie.append([order_data[0], None])
                return True
        print("Order ID not found %s ... \n" % str(order_data))

    def plot(self, filename='index.html'):
        dir_name = str(random.randint(0, 100))
        os.chdir(System.get_cache_dir())
        dir_list = os.listdir()
        if dir_name not in dir_list:
            os.mkdir(dir_name)
        self.save_to_file("/"+dir_name+"/")
        WebServer.serve(
            routing={
                "/data/": "%s/%s/" % (System.get_cache_dir(), dir_name),
                "/": "%s/lib" % os.path.dirname(__file__),
            },
            open_browser=True,
            timeout=40,
            filename='/' + filename
        )
        shutil.rmtree(dir_name)

    def save_to_file(self, dir_name):
        plot_data = {"buy": self.series['buy']['data'], "sell": self.series['sell']['data'], "assets": self.series['assets']}
        logging.info("Buy orders count: %s" % str(len(plot_data['buy'])))
        logging.info("Sell orders count: %s" % str(len(plot_data['sell'])))
        with open('%s%sorders_data.json' % (System.get_cache_dir(), dir_name), 'w') as outfile:
            json.dump(self.series, outfile)

    def new_price(self):
        time.sleep(1)
        data = {"OHLC": self.series["OHLC"]['data'][-1],
                "volume": self.series["volume"]['data'][-1],
                "pos_vol": self.series["position"]['data'][-1],
                "equity": self.series["equity"]['data'][-1],
                "base_equity": self.series["base_equity"]['data'][-1],
                "hold": self.series["hold"]['data'][-1],
                "pos_p": self.series["position_price"]['data'][-1]
                }
        balance_data = {
                "pos_vol": self.series["position"]['data'][-1],
                "equity": self.series["equity"]['data'][-1],
                "base_equity": self.series["base_equity"]['data'][-1],
                "hold": self.series["hold"]['data'][-1]
                }

        asyncio.get_event_loop().run_until_complete(self.update_plot({"action": "produce",
                                                                      "type": "price",
                                                                      "data": data
                                                                      }, self.chart_port))
        asyncio.get_event_loop().run_until_complete(self.update_plot({"action": "produce",
                                                                      "type": "price",
                                                                      "data": balance_data
                                                                      }, self.balance_port))

    async def update_plot(self, data, port):
        uri = "ws://localhost:" + str(port)
        async with websockets.connect(uri) as websocket:
            await websocket.send(json.dumps(data))
            await websocket.close()



import setuptools
setuptools.setup(
      name="tradingkit",
      version="1.9.3",
      author="QBit Artifacts, SL",
      author_email="lluis@logictraders.com",
      license="MIT",
      description="Trading and backtesting framework for Python",
      long_description=open("README.md", "r").read(),
      long_description_content_type="text/markdown",
      url="https://github.com/logictraders/tradingkit",
      packages=setuptools.find_namespace_packages(where="src", include=['tradingkit.*']),
      package_dir={"": "src"},
      include_package_data=True,
      entry_points={
            "console_scripts": [
                  "tk = tradingkit.cli.cli:CLI.main"
            ]
      },
      classifiers=[
            "Development Status :: 3 - Alpha",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Environment :: Console",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Intended Audience :: Developers",
            "Topic :: Office/Business",
            "Topic :: Software Development",
            "Topic :: Software Development :: Build Tools",
            "Topic :: Software Development :: Libraries",
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: Software Development :: Libraries :: Application Frameworks"
      ],
      install_requires=[line.strip() for line in open("requirements.txt").readlines() if len(line) > 1]
)




import json
import logging
from datetime import datetime

from dateutil.relativedelta import relativedelta

from tradingkit.utils.system import System
from tradingkit.data.feed.feeder import Feeder
from tradingkit.pubsub.core.publisher import Publisher
from tradingkit.pubsub.event.trade import Trade
from ccxt import InsufficientFunds


class BacktestFeeder(Feeder, Publisher):
    def __init__(self, exchange='kraken', symbol='BTC/EUR', since8601=None, to8601=None):
        super().__init__()
        self.since = datetime.fromisoformat(since8601)
        self.to = datetime.fromisoformat(to8601)
        self.symbol = symbol
        self.exchange = exchange

    def feed(self):
        start_month = self.since.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        end_month = self.to.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        while start_month <= end_month:
            try:
                self.dispatch_month(
                    self.exchange,
                    self.symbol,
                    start_month.year,
                    start_month.month,
                    self.since,
                    self.to
                )
                start_month += relativedelta(months=1)
            except FileNotFoundError as e:
                if start_month.timestamp() < datetime.now().timestamp():
                    logging.warning("Import file not found, use command 'tk import' before")
                    raise e
                break

            except InsufficientFunds as e:
                print("InsufficientFunds: Executed Liquidation")
                break

    def dispatch_month(self, exchange, symbol, year, month, since, to):
        import_dir = System.get_import_dir()
        base, quote = symbol.split('/')
        full_filename = '%s/%s-%s_%s-%d-%02d.json' % (import_dir, exchange, base, quote, year, month)

        for trade in json.load(open(full_filename, 'r')):
            if since.timestamp() <= trade['timestamp'] / 1000 < to.timestamp():
                self.dispatch(Trade(trade))



import hashlib
import hmac
import inspect
import sys
import time
import json
import traceback
import logging

import ccxt
from ccxt import NetworkError
from dateutil import parser
from functools import partial

import websocket
from websocket import _logging

from tradingkit.data.feed.feeder import Feeder
from tradingkit.data.feed.websocket_feeder import WebsocketFeeder
from tradingkit.pubsub.core.publisher import Publisher
from tradingkit.pubsub.event.book import Book
from tradingkit.pubsub.event.funding import Funding
from tradingkit.pubsub.event.order import Order
from tradingkit.pubsub.event.position import Position
from tradingkit.pubsub.event.trade import Trade


class BitmexFeeder(WebsocketFeeder):

    BITMEX_SYMBOL_MAP = {
        'BTC/USD': 'XBTUSD',
        'BTC/USDT': 'XBTUSDT'
    }

    BITMEX_SYMBOL_MAP_REV = {
        'XBTUSD': 'BTC/USD',
        'XBTUSDT': 'BTC/USDT'
    }

    def __init__(self, symbol='BTC/USD', credentials=None, url="wss://ws.bitmex.com/realtime"):
        super().__init__(symbol, credentials, url)

    def on_open(self, ws):
        self.subscribe(ws, 'trade:%s' % self.BITMEX_SYMBOL_MAP[self.symbol])
        self.subscribe(ws, 'orderBook10:%s' % self.BITMEX_SYMBOL_MAP[self.symbol])
        self.subscribe(ws, 'funding:%s' % self.BITMEX_SYMBOL_MAP[self.symbol])

        if self.credentials is not None:
            self.authenticate(ws)
            self.subscribe(ws, 'order')
            self.subscribe(ws, 'position')

    def authenticate(self, ws):
        nonce = int(time.time()) + 100
        message = 'GET/realtime' + str(nonce)
        signature = hmac.new(
            self.credentials['secret'].encode('utf-8'),
            message.encode('utf-8'),
            digestmod=hashlib.sha256
        ).hexdigest()
        ws.send(json.dumps({'op': 'authKeyExpires', 'args': [self.credentials['apiKey'], nonce, signature]}))

    def subscribe(self, ws, topic):
        ws.send(json.dumps({'op': 'subscribe', 'args': topic}))

    def on_message(self, ws, message):
        payload = json.loads(message)
        if 'table' in payload:
            if payload['table'] == 'orderBook10' and payload['data']:
                order_book = self.transform_book_data(payload)
                if order_book is not None:
                    self.dispatch(Book(order_book))

            elif payload['table'] == 'trade' and payload['data']:
                trade = self.transform_trade_data(payload)
                self.dispatch(Trade(trade))

            elif payload['table'] == 'order' and payload['data']:
                if 'ordStatus' in payload['data'][0].keys() and payload['data'][0]['ordStatus'] == 'Filled':
                    order_data = self.transform_order_data(payload)
                    self.dispatch(Order(order_data))

            elif payload['table'] == 'position' and payload['data']:
                self.dispatch(Position(payload['data'][0]))

            elif payload['table'] == 'funding' and payload['data']:
                self.dispatch(Funding(payload['data'][0]))

            else:
                print("Unknown table Message:", str(payload))
        else:
            print("Unknown Message:", str(payload))

    def transform_book_data(self, payload):
        order_book = payload['data'][0]
        order_book['timestamp'] = int(parser.isoparse(payload['data'][0]['timestamp']).timestamp() * 1000)
        order_book['symbol'] = self.BITMEX_SYMBOL_MAP_REV[order_book['symbol']]

        return order_book

    def transform_trade_data(self, payload):
        trade = payload['data'][0].copy()
        trade['timestamp'] = parser.isoparse(trade['timestamp']).timestamp() * 1000
        trade['symbol'] = self.BITMEX_SYMBOL_MAP_REV[trade['symbol']]

        trade['amount'] = trade['size']
        trade['info'] = payload['data'][0].copy()
        if 'cost' not in trade.keys():
            trade['cost'] = float(trade['size']) * float(trade['price'])
        return trade

    def transform_order_data(self, payload):
        symbol = self.BITMEX_SYMBOL_MAP_REV[payload['data'][0]['symbol']]
        timestamp = int(parser.isoparse(payload['data'][0]['timestamp']).timestamp() * 1000)
        logging.debug("PAYLOAD: %s" % str(payload))

        order_payload = {
            "info": payload['data'][0].copy(),
            "id": payload['data'][0]['orderID'],
            "status": payload['data'][0]['ordStatus'].lower(),
            "amount": payload['data'][0]['cumQty'],
            "timestamp": timestamp,
            "lastTradeTimestamp": int(time.time() * 1000),
            "symbol": symbol,
            "leavesQty": payload['data'][0]['leavesQty']
        }

        # sometimes bitmex order updates doesn't have avgPx
        if 'avgPx' in payload['data'][0]:
            order_payload["price"] = payload['data'][0]['avgPx']
        return order_payload



from tradingkit.exchange.testex import TestEX
from datetime import datetime

from tradingkit.pubsub.core.event import Event
from tradingkit.pubsub.event.funding import Funding
from tradingkit.pubsub.event.liquidation import Liquidation
from tradingkit.pubsub.event.plot import Plot
import logging

from tradingkit.pubsub.event.trade import Trade
from ccxt import InsufficientFunds


class BitmexBacktest(TestEX):
    def sleep(self, milliseconds):
        pass

    def __init__(self, params=None):
        super().__init__({'balance': {'USD': 0, 'BTC': 10}, 'fees': {'maker': -0.00025, 'taker': 0.00075}})
        self.position = {"currentQty": 0,  # currentQty
                         "homeNotional": 0,  # homeNotional base value (disabled)
                         "avgEntryPrice": 0,
                         "liquidationPrice": None,
                         "openOrderBuyQty": None,
                         "openOrderSellQty": None,
                         "lastPrice": None,
                         "markPrice": None
                         }
        self.leverage = 100
        self.funding_rate = None

    def match_order(self, trade, order, price, base, quote):
        # TODO:
        #   - partial fills
        fee = self.fees['taker'] if order['type'] == 'market' else self.fees['maker']
        if order['side'] == 'buy':
            if order['type'] == 'stop':
                if price <= trade['price']:
                    self.update_position(order['amount'], trade['price'], base)
                    order['status'] = 'filled'
            elif price >= trade['price']:
                self.update_position(order['amount'], price, base)
                order['status'] = 'filled'
        else:
            if order['type'] == 'stop':
                if price >= trade['price']:
                    self.update_position(-order['amount'], trade['price'], base)
                    order['status'] = 'filled'
            elif price <= trade['price']:
                self.update_position(-order['amount'], price, base)
                order['status'] = 'filled'
        if order['status'] == 'filled':
            order['lastTradeTimestamp'] = trade['timestamp']
            if order['price'] is None:
                order['price'] = price
            fee = self.fees['maker'] if order['type'] == 'limit' else self.fees['taker']
            fee_pnl = abs(order['amount']) * fee
            logging.debug("fees apply: %s $" % str(-fee_pnl))
            self.balance[base] -= fee_pnl / price
            self.orders_scheduled_to_close.append(order)
            self.plot_balances(base, quote, price)

    def update_position(self, quote_volume, avg_order_price, base):
        against = quote_volume * self.position['currentQty'] < 0
        if against:
            if abs(self.position['currentQty']) < abs(quote_volume):  # position changes sign, update PNL
                pnl = (avg_order_price / self.position['avgEntryPrice'] - 1) * self.position['currentQty']
                self.balance[base] += pnl / avg_order_price
                self.position['avgEntryPrice'] = avg_order_price

            else:  # decrease position, update PNL
                pnl = (avg_order_price / self.position['avgEntryPrice'] - 1) * abs(quote_volume)
                if self.position['currentQty'] >= 0:
                    self.balance[base] += pnl / avg_order_price
                else:
                    self.balance[base] -= pnl / avg_order_price
        else:  # increase position, update position price (mean price)
            self.position['avgEntryPrice'] = (self.position['currentQty'] * self.position['avgEntryPrice'] +
                                              quote_volume * avg_order_price) / (
                                                     self.position['currentQty'] + quote_volume)
        self.position['currentQty'] += quote_volume
        self.set_liquidation_price(avg_order_price, base)

        logging.debug("balance: %s" % str(self.balance))
        logging.debug("position_updated: %s" % str(self.position))

    def on_event(self, event: Event):
        if isinstance(event, Trade):
            self.set_mark_price(event.payload)
            if self.position['currentQty'] > 0:
                if self.position['liquidationPrice'] >= self.position['markPrice']:
                    self.execute_liquidation(event.payload)

            elif self.position['currentQty'] < 0:
                if self.position['liquidationPrice'] <= self.position['markPrice']:
                    self.execute_liquidation(event.payload)

        if isinstance(event, Funding):
            base, quote = event.payload['symbol'].split('/')
            self.balance[base] -= self.position['currentQty'] * event.payload['rate'] / event.payload['price']
            self.funding_rate = {"timestamp": event.payload['timestamp'], "rate": event.payload['rate']}
        super().on_event(event)

    def execute_liquidation(self, trade):
        time = datetime.fromtimestamp(self.seconds())
        markPrice = self.position['markPrice']
        base = trade['symbol'].split('/')[0]
        price = trade['price']
        free_balance = self.fetch_free_balance()[base] * price
        max_leverage = 200
        leverage = min(abs(self.position['currentQty']) / free_balance, max_leverage) if free_balance > 0 else max_leverage
        liquidationPrice = round(self.position['liquidationPrice'], 2)
        logging.warning("InsufficientFunds: Executed Liquidation\n\tTime: %s\n\tMarket Price: %s\n\tLeverage: %s\n\t"
                     "Liquidation Price: %s\n\t" % (str(time), str(markPrice), str(leverage), str(liquidationPrice)))

        self.open_orders = {}
        for asset in self.balance.keys():
            self.balance[asset] = 0
        self.position = {"currentQty": 0,
                         "homeNotional": 0,
                         "avgEntryPrice": 0,
                         "liquidationPrice": None,
                         "openOrderBuyQty": None,
                         "openOrderSellQty": None,
                         "lastPrice": None,
                         "markPrice": None
                         }
        self.dispatch(Liquidation(trade))
        raise InsufficientFunds("Zero balance: %s" % str(self.balance))

    def plot_balances(self, base, quote, price):
        exchange_date = datetime.fromtimestamp(self.timestamp / 1000.0).isoformat()
        value = self.balance[quote] + self.balance[base]
        quote_value = value * price
        self.dispatch(Plot({
            'name': 'Equity',
            'type': 'scatter',
            'mode': 'lines',
            'color': 'blue',
            'yaxis': 'balance',
            'data': {
                'x': exchange_date,
                'y': quote_value,
                'base_value': value,
                'position_vol': self.position['currentQty'],
                'position_price': self.position['avgEntryPrice'],
                'tooltip': "%.2f %s" % (quote_value, quote)
            },
        }))

    def funds_enough(self, order):
        sum_same_side_orders = sum([o['amount'] for o in self.open_orders.values() if o['side'] == order['side']])
        base, quote = order['symbol'].split('/')
        price = self.orderbooks[order['symbol']]['bids'][0][0]
        max_balance_avalable = self.balance[base] * price * self.leverage - abs(
            self.position["currentQty"]) - sum_same_side_orders
        return max_balance_avalable - order['amount'] >= 0

    def private_get_position(self):
        return [self.position]

    def set_mark_price(self, trade):
        """
        bitmex formula for markPrice/FairPrice
        FundingBasis = FundingRate * (Time Until Funding / Funding Interval)
        FairPrice = IndexPrice * (1 + Funding Basis)
        """
        if self.funding_rate is None:  # compatibility with no funding feeder
            self.position['markPrice'] = trade['price']
        else:
            time_until_founding = self.funding_rate['timestamp'] + 28800000 - trade['timestamp']
            founding_basis = self.funding_rate['rate'] * (time_until_founding / 28800000)
            self.position['markPrice'] = trade['price'] * (1 + founding_basis)

    def cancel_order(self, order_id, symbol=None, params={}):
        if order_id in self.open_orders:
            symbol = self.open_orders[order_id]['symbol']
            base = symbol.split('/')[0]
            mark_price = self.fetch_ticker(symbol)['bid']
            super().cancel_order(order_id, symbol, params)
            self.set_liquidation_price(mark_price, base)
        else:
            super().cancel_order(order_id, symbol, params)

    def create_order(self, symbol, type, side, amount, price=None, params={}):
        response = super().create_order(symbol, type, side, amount, price, params)
        base = symbol.split('/')[0]
        mark_price = self.fetch_ticker(symbol)['bid']
        self.set_liquidation_price(mark_price, base)
        return response

    def get_liquidation_price(self, mark_price, base):
        if abs(self.position['currentQty']) > 0:
            side = 'buy' if self.position['currentQty'] > 0 else 'sell'
            sum_same_side_orders = sum(
                [(o['amount'] / o['price']) for o in self.open_orders.values() if o['side'] == side and o['type'] == 'limit'])
            free_base_balance = self.balance[base] - sum_same_side_orders
            leverage_ratio = (free_base_balance / self.position['currentQty'] + 1.0 / self.position['avgEntryPrice'])
            return 1.0 / leverage_ratio if leverage_ratio > 0 else 100000000.0  # bitmex max liquidation price
        return 0

    def set_liquidation_price(self, mark_price, base):
        self.position['liquidationPrice'] = self.get_liquidation_price(mark_price, base)

    def get_liquidation_price_old(self, mark_price, base):
        if abs(self.position['currentQty']) > 0:
            side = 'buy' if self.position['currentQty'] > 0 else 'sell'
            sum_same_side_orders = sum([o['amount'] for o in self.open_orders.values() if o['side'] == side])
            if self.position['currentQty'] < 0:
                sum_same_side_orders = -sum_same_side_orders

            price = list(self.orderbooks.values())[0]['bids'][0][0]

            free_base_balance = self.balance[base] + 0
            if abs(self.position['currentQty']) > 0:
                pnl = (price / self.position['avgEntryPrice'] * self.position['currentQty'] -
                       self.position['currentQty']) / price
                free_base_balance += pnl

            leverage = max(abs(self.position['currentQty']) / (free_base_balance * price), 1)
            available_margin = self.balance[base] - abs(sum_same_side_orders / leverage) / price

            bankruptcy_price = 1 / (1 / self.position['avgEntryPrice'] + available_margin / self.position['currentQty'])
            bankruptcy_value = self.position['currentQty'] * (1 / bankruptcy_price)
            funding_rate = self.funding_rate['rate'] if self.funding_rate else 0
            maintenance_margin = (0.005 * (self.position['currentQty'] / self.position['avgEntryPrice'])) + \
                                 (0.00075 * bankruptcy_value) + \
                                 (0.0001 * funding_rate)

            lp = 1 / (
                        1 / self.position['avgEntryPrice'] + (available_margin - maintenance_margin) / self.position[
                    'currentQty'])
            if lp < 0:
                return 100000000.0  # bitmex max liquidation price
            return lp
        return None

    def fetch_balance(self):
        balances = {}
        symbol = list(self.orderbooks.keys())[0]
        price = list(self.orderbooks.values())[0]['bids'][0][0]
        base, quote = symbol.split('/')
        free_balance = self.fetch_free_balance()
        balances['free'] = free_balance
        used_balance = self.fetch_used_balance()
        balances['total'] = self.balance.copy()
        for asset in used_balance.keys():
            if asset == base:
                total = self.balance[asset]
                balances[asset] = {'free': free_balance[asset], 'used': used_balance[asset], 'total': total}
                balances['total'][asset] = total
                balances['info'] = [{'walletBalance': self.balance[asset] * 1e8}]
            else:
                balances[asset] = {'free': free_balance[asset], 'used': used_balance[asset],
                                   'total': self.balance[asset]}
        balances['used'] = used_balance
        return balances

    def fetch_free_balance(self):
        free_balance = self.balance.copy()
        base, quote = list(self.orderbooks.keys())[0].split('/')
        price = list(self.orderbooks.values())[0]['bids'][0][0]

        if abs(self.position['currentQty']) > 0:
            pnl = (price / self.position['avgEntryPrice'] * self.position['currentQty'] -
                   self.position['currentQty']) / price
            free_balance[base] += pnl

        if len(self.open_orders) > 0:
            side = 'buy' if self.position['currentQty'] > 0 else 'sell'
            sum_same_side_orders = sum(
                [(o['amount'] / o['price']) for o in self.open_orders.values() if o['side'] == side])
            leverage = max(abs(self.position['currentQty']) / (free_balance[base] * price), 1)
            free_balance[base] -= sum_same_side_orders / leverage

        return free_balance



import functools
import logging
import random
import socketserver
import time
from multiprocessing.context import Process

from tradingkit.utils.request_handler import RequestHandler


class WebServer:

    @staticmethod
    def serve(routing: dict, open_browser=False, timeout=None, filename=''):
        p = Process(target=WebServer.serve_and_browse, args=(routing, open_browser, filename, ))
        p.start()
        wait_for_server_seconds = timeout
        logging.info("Waiting for server %d seconds" % wait_for_server_seconds)
        time.sleep(wait_for_server_seconds)
        p.terminate()

    @staticmethod
    def serve_and_browse(routing: dict, open_browser=False, filename=''):
        handler = functools.partial(RequestHandler, routing)
        port = random.randint(1024, 65535)
        with socketserver.TCPServer(("", port), handler) as httpd:
            if open_browser:
                import webbrowser
                webbrowser.open_new_tab('http://localhost:%d%s' % (port, filename))
            print("serving at port", port)
            httpd.serve_forever()



from datetime import datetime, timedelta

from ccxt import Exchange

from tradingkit.pubsub.core.event import Event
from tradingkit.pubsub.core.publisher import Publisher
from tradingkit.pubsub.event.book import Book
from tradingkit.pubsub.core.subscriber import Subscriber
from tradingkit.pubsub.event.candle import Candle
from tradingkit.pubsub.event.funding import Funding
from tradingkit.pubsub.event.open_order import OpenOrder
from tradingkit.pubsub.event.order import Order
from tradingkit.pubsub.event.plot import Plot
from tradingkit.pubsub.event.trade import Trade
from tradingkit.pubsub.event.liquidation import Liquidation
import numpy


class BridgeExchange(Publisher, Subscriber, Exchange):

    def __init__(self, exchange: Exchange):
        super().__init__()
        self.exchange = exchange
        self.closed_orders = {}
        self.orders_history = {}
        self.balance_history = []
        self.last_balance_check = None
        self.peak_balance = 0
        self.max_drawdown = 0
        self.last_price = None
        self.symbol = None
        self.has_position = True if "bitmex" in str(exchange.__class__) else False
        self.is_backtest = True if "TestEX" in str(exchange.__class__) or "BitmexBacktest" in str(
            exchange.__class__) else False
        exchange.seconds()

        self.candles = None
        self.last_candle = None
        self.timeframes = {
            "1m": 60,
            "5m": 300,
            "15m": 900,
            "30m": 1800,
            "1h": 3600,
            "4h": 14400,
            "6h": 21600,
            "12h": 43200,
            "1d": 86400,
            "2d": 172800,
            "4d": 345600,
            "1w": 604800,
            "2w": 1209600
        }

    def sec(self):
        return self.exchange.sec()

    def msec(self):
        return self.exchange.msec()

    def usec(self):
        return self.exchange.usec()

    def seconds(self):
        return self.exchange.seconds()

    def milliseconds(self):
        return self.exchange.milliseconds()

    def microseconds(self):
        return self.exchange.microseconds()

    def subscribed_events(self) -> list:
        return [Order, Trade, Book, Candle, Liquidation, Funding, OpenOrder]

    def get_order_asset(self, symbol):
        base, quote = symbol.split('/')
        return quote if self.is_bitmex() else base

    def is_bitmex(self):
        exchange_name = str(self.exchange.__class__.__name__).lower()
        return "bitmex" in exchange_name

    def is_inverse(self, symbol):
        return self.is_bitmex()

    def max_leverage(self, symbol=None):
        return 100 if self.is_bitmex() else 1

    def on_event(self, event: Event):
        if isinstance(event, Book):
            if self.last_price is None:
                self.calculate_exchange_state(event.payload['timestamp'], event.payload['symbol'], event.payload['bids'][0][0])
                self.symbol = event.payload['symbol']
            self.last_price = event.payload['bids'][0][0]
        if isinstance(event, Candle):
            self.plot_candle(event)
            if self.is_backtest:
                self.update_balance_hist(event)
        if isinstance(event, Order):
            order = event.payload.copy()
            if order['id'] in self.orders_history.keys():
                self.orders_history[order['id']].update(order)
                event.payload = self.orders_history[order['id']]
            self.plot_order(event)
            self.calculate_exchange_state(order['lastTradeTimestamp'], order['symbol'], self.last_price)
        if isinstance(event, OpenOrder):
            self.plot_order(event)
            order = event.payload.copy()
            self.calculate_exchange_state(order['timestamp'], order['symbol'], self.last_price)
        if isinstance(event, Liquidation):
            trade = event.payload
            self.calculate_exchange_state(trade['timestamp'], trade['symbol'], trade['price'])
        if isinstance(event, Trade):
            trade = event.payload
            if self.symbol is None:
                self.symbol = event.payload['symbol']
            self.candle_dispatcher(trade)
        self.dispatch(event)

    def fetch_open_orders(self, symbol=None, since=None, limit=None, params={}):
        return self.exchange.fetch_open_orders(symbol, since, limit, params)

    def create_order(self, symbol, type, side=None, amount=0, price=None, params={}):
        if amount == 0:
            raise ValueError("Zero order amount is not allowed!!!")
        elif amount < 0 and side is not None:
            raise ValueError("Negative order amount is not allowed!!!")
        if side is None:
            side = "buy" if amount > 0 else "sell"
            if amount < 0:
                amount = -amount
        order = self.exchange.create_order(symbol, type, side, amount, price, params)
        self.orders_history[order['id']] = order
        return order

    def cancel_order(self, order_id, symbol=None, params={}):
        return self.exchange.cancel_order(order_id, symbol, params)

    def match_order(self, trade, order, price, base, quote):
        return self.exchange.match_order(trade, order, price, base, quote)

    def fetch_ticker(self, symbol, **kwargs):
        return self.exchange.fetch_ticker(symbol)

    def private_get_position(self):
        if self.has_position:
            position = self.exchange.private_get_position()
            if not position:
                return [{"currentQty": 0,
                         "homeNotional": 0,
                         "avgEntryPrice": 0,
                         "liquidationPrice": None,
                         "openOrderBuyQty": None,
                         "openOrderSellQty": None,
                         "lastPrice": None,
                         "markPrice": None
                         }]
            else:
                return position
        else:
            raise ValueError("Position not implemented only bitmex !!!")

    def fetch_balance(self):
        return self.exchange.fetch_balance()

    def fetch_free_balance(self):
        return self.exchange.fetch_free_balance()

    def fetch_used_balance(self):
        return self.exchange.fetch_used_balance()

    def fetch_total_balance(self):
        return self.exchange.fetch_total_balance()

    def fetchMarkets(self):
        return self.exchange.fetchMarkets()

    def get_max_draw_down(self):
        return self.max_drawdown

    def getPairs(self, symbol):
        return symbol

    def plot_balances(self, exchange_state):
        quote = exchange_state['quote']
        base = exchange_state['base']

        ttip = "<b>Total: %.2f %s</b><br />Quote: %.2f %s<br />Base: %.2f %s (in %s)" % (
            exchange_state['equity'],
            quote,
            exchange_state['quote_balance'],
            exchange_state['quote'],
            exchange_state['base_balance'] * exchange_state['price'],
            quote,
            base
        )
        self.dispatch(Plot({
            'name': 'Equity',
            'type': 'scatter',
            'mode': 'lines',
            'color': 'blue',
            'yaxis': 'balance',
            'quote': quote,
            'base': base,
            'data': {
                'x': exchange_state['exchange_date'],
                'y': exchange_state['equity'],
                'base_equity': exchange_state['base_equity'],
                'base_balance': exchange_state['all_balances']['total'][base],
                'quote_balance': exchange_state[
                    'all_balances']['total'][quote] if quote in exchange_state['all_balances'][ 'total'] else 0,
                'position_vol': exchange_state['position_vol'],
                'position_price': exchange_state['position_price'],
                'invested': exchange_state['base_balance'] * exchange_state['price'],
                'tooltip': ttip
            },
        }))

    def calculate_exchange_state(self, price, symbol, timestamp):
        exchange_date = datetime.fromtimestamp(timestamp / 1000.0).isoformat()
        base, quote = symbol.split('/')
        all_balances = self.fetch_balance()
        balances = all_balances['free'] if all_balances['free'][base] else all_balances['total']
        base_balance = balances[base] if base in balances else 0
        quote_balance = balances[quote] if quote in balances else 0
        equity = quote_balance + base_balance * price
        base_equity = base_balance + quote_balance / price
        if self.has_position:
            position = self.private_get_position()[0]
            position_vol = position['currentQty']
            position_price = position['avgEntryPrice']

        else:
            position_vol = base_balance * price
            position_price = 0

        self.plot_balances({
            'all_balances': all_balances,
            'base_balance': base_balance,
            'base_equity': base_equity,
            'equity': equity,
            'exchange_date': exchange_date,
            'position_price': position_price,
            'position_vol': position_vol,
            'quote': quote,
            'base': base,
            'price': price,
            'quote_balance': quote_balance
        })
        if self.is_backtest:
            self.calculate_max_drawdown(base_balance, quote_balance)

    def plot_order(self, event):
        order = event.payload
        name = None
        if order['status'].lower() == 'filled':
            fill_date = datetime.fromtimestamp(order['lastTradeTimestamp'] / 1000.0).isoformat()
            name = 'buy' if order['side'] == 'buy' else 'sell'
        else:
            fill_date = datetime.fromtimestamp(order['timestamp'] / 1000.0).isoformat()
            if order['status'].lower() == 'open':
                name = 'open_buy' if order['side'] == 'buy' else 'open_sell'
            elif order['status'].lower() == 'canceled':
                name = 'cancel_buy' if order['side'] == 'buy' else 'cancel_sell'
        if name:
            self.dispatch(Plot({
                'name': name,
                'type': 'scatter',
                'mode': 'markers',
                'color': 'green' if order['side'] == 'buy' else 'red',
                'yaxis': 'price',
                'data': {
                    'x': fill_date,
                    'y': order['price'],
                    'type': order['type'],
                    'id': order['id'],
                    'tooltip': "%.4f@%.2f" % (
                        order['amount'] if order['side'] == 'buy' else -order['amount'],
                        order['price']
                    )
                },
            }))

    def plot_candle(self, event):
        candle = event.payload
        candle['liquidationPrice'] = self.private_get_position()[0]['liquidationPrice'] if self.has_position else None
        self.dispatch(Plot({
            'name': 'Price',
            'type': 'candlestick',
            'yaxis': 'price',
            'data': candle,
        }))

    def calculate_max_drawdown(self, base_balance, quote_balance):
        if self.last_price:
            balance = quote_balance + base_balance * self.last_price if quote_balance > 0 else base_balance

            if balance > self.peak_balance:
                self.peak_balance = balance

            drawdown = (balance - self.peak_balance) / self.peak_balance
            self.max_drawdown = min(self.max_drawdown, drawdown)

    def update_balance_hist(self, event):
        date = datetime.fromisoformat(event.payload['datetime'])
        if self.last_balance_check is None or date - self.last_balance_check > timedelta(hours=1):
            self.last_balance_check = date

            price = event.payload['close']
            base, quote = self.symbol.split('/')
            balances = self.fetch_balance()
            balances = balances['free'] if balances['free'][base] else balances['total']

            base_balance = balances[base] if base in balances else 0
            quote_balance = balances[quote] if quote in balances else 0

            # todo use quote equity for kraken (equity = quote_balance + base_balance * price)
            base_equity = base_balance + quote_balance / price

            self.balance_history.append(base_equity)

    def get_sharpe_ratio(self):
        standard_deviation = numpy.std(self.balance_history)
        sharpe_ratio = (self.balance_history[-1] / self.balance_history[0]) / standard_deviation
        return sharpe_ratio

    def candle_dispatcher(self, trade):

        if self.candles is None:
            self.candles = {x: {} for x in self.timeframes.keys()}

        if self.last_candle is None:
            self.last_candle = {x: None for x in self.timeframes.keys()}

        for tf in self.timeframes.keys():
            sec = self.timeframes[tf]
            key = trade['timestamp'] // (sec * 1000) * sec
            key = str(datetime.fromtimestamp(key))
            if self.last_candle[tf] is not None and key in self.candles[tf]:
                self.candles[tf][key]['high'] = max(self.candles[tf][key]['high'], trade['price'])
                self.candles[tf][key]['low'] = min(self.candles[tf][key]['low'], trade['price'])
                self.candles[tf][key]['close'] = trade['price']
                self.candles[tf][key]['vol'] += trade['amount']
                self.candles[tf][key]['cost'] += trade['cost']
                self.candles[tf][key]['trades'] += 1
            else:
                self.candles[tf][key] = {
                    'datetime': key,
                    'open': trade['price'],
                    'high': trade['price'],
                    'low': trade['price'],
                    'close': trade['price'],
                    'vol': trade['amount'],
                    'cost': trade['cost'],
                    'trades': 1,
                    'timeframe': tf
                }
                if self.last_candle[tf] is not None:
                    candle = Candle(self.last_candle[tf])
                    self.dispatch(candle)
                    self.plot_candle(candle)
                    if self.is_backtest:
                        self.update_balance_hist(candle)
            self.last_candle[tf] = self.candles[tf][key]



import plotly.graph_objects as go
import pandas as pd

from tradingkit.display.plotter import Plotter
from tradingkit.pubsub.core.event import Event
from tradingkit.pubsub.event.plot import Plot


class PlotlyPlotter(Plotter):

    def __init__(self):
        self.price = []
        self.series = {}

    def on_event(self, event: Event):
        if isinstance(event, Plot):
            plot = event.payload
            if plot['name'].lower() == 'price':
                self.price.append(plot['data'])
            else:
                if plot['name'] not in self.series:
                    self.series[plot['name']] = plot.copy()
                    self.series[plot['name']]['data'] = []
                data = plot['data'].copy()
                if 'tooltip' not in data:
                    data['tooltip'] = data['y']
                self.series[plot['name']]['data'].append(data)

    def plot(self):

        df = pd.DataFrame(self.price)
        if len(df) > 10000:
            data = [{
                'type': 'scatter',
                'name': 'Price',
                'yaxis': 'y2',
                'x': df['datetime'],
                'y': df['close']
            }]
        else:
            data = [{
                'type': 'candlestick',
                'name': 'Price',
                'yaxis': 'y2',
                'x': df['datetime'],
                'open': df['open'],
                'high': df['high'],
                'low': df['low'],
                'close': df['close']
            }]

        for serie in self.series.values():
            s = pd.DataFrame(serie['data'])
            data.append({
                'name': serie['name'],
                'type': serie['type'],
                'mode': serie['mode'],
                'line_color': serie['color'],
                'fillcolor': serie['color'],
                'x': s['x'],
                'y': s['y'],
                'text': s['tooltip'],
                'hovertemplate': '<b>%{text}</b>',
                'yaxis': 'y2' if serie['yaxis'] == 'price' else 'y',
                'line_width': 2
            })

        layout = {
            'xaxis': {'rangeselector': {'visible': True}},
            'yaxis': {'domain': [0, 0.1]},
            'yaxis2': {'domain': [0.1, 1]}
        }

        fig = go.Figure({'data': data, 'layout': layout})

        fig.show()

